from typing import Optional

from pyasn1.codec.der.encoder import encode
from pyasn1.type import univ, base
from pyasn1_alt_modules import rfc3279, rfc5280, rfc5480, rfc5639, rfc4055

from pkilint import validation
from pkilint.pkix import algorithm
from pkilint.pkix.certificate import certificate_key


class RsaKeyValidator(validation.Validator):
    """
    TS 119 312, clause 6.2.2.1:

    The public exponent e shall be an odd positive integer such that 2^16 < e < 2^256.
    """

    VALIDATION_RSA_EXPONENT_OUT_OF_RANGE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "ts_119_312.6.2.2.1.rsa_exponent_of_range",
    )

    VALIDATION_RSA_SMALL_MODULUS = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE, "ts_119_312.8.4.rsa_small_modulus"
    )

    _MIN_MODULUS_LENGTH = 1900
    _MIN_EXPONENT_EXCLUSIVE = 1 << 16
    _MAX_EXPONENT_EXCLUSIVE = 1 << 256

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_RSA_EXPONENT_OUT_OF_RANGE,
                self.VALIDATION_RSA_SMALL_MODULUS,
            ],
            pdu_class=rfc3279.RSAPublicKey,
        )

    def validate(self, node):
        modulus_len = int(node.children["modulus"].pdu).bit_length()
        exponent_int = int(node.children["publicExponent"].pdu)

        findings = []

        if modulus_len < 1900:
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_RSA_SMALL_MODULUS,
                    f"RSA public key has a modulus length of {modulus_len} bits",
                )
            )

        if (
            not self._MIN_EXPONENT_EXCLUSIVE
            < exponent_int
            < self._MAX_EXPONENT_EXCLUSIVE
        ):
            findings.append(
                validation.ValidationFindingDescription(
                    self.VALIDATION_RSA_EXPONENT_OUT_OF_RANGE,
                    f"RSA public key has an exponent of {exponent_int}",
                )
            )

        return validation.ValidationResult(self, node, findings)


def _create_alg_id_der(
    o: univ.ObjectIdentifier, params: Optional[base.Asn1Type]
) -> bytes:
    alg_id = rfc5280.AlgorithmIdentifier()
    alg_id["algorithm"] = o

    if params is not None:
        alg_id["parameters"] = params

    return encode(alg_id)


_SPKI_ALG_ID_ENCODINGS = (
    [
        # RSA
        _create_alg_id_der(rfc5480.rsaEncryption, univ.Null(""))
    ]
    + [
        # ECDSA
        _create_alg_id_der(rfc5480.id_ecPublicKey, c)
        for c in (
            univ.ObjectIdentifier("1.2.250.1.223.101.256.1"),  # FRP256v1
            rfc5639.brainpoolP256r1,
            rfc5639.brainpoolP384r1,
            rfc5639.brainpoolP512r1,
            rfc5480.secp256r1,
            rfc5480.secp384r1,
            rfc5480.secp521r1,
        )
    ]
    # TODO: add DSA
)


class AllowedPublicKeyTypeValidator(
    certificate_key.AllowedPublicKeyAlgorithmEncodingValidator
):
    """
    From EN 319 412 2:

    GEN-4.2.5-1

    The subject public key should be selected according to ETSI TS 119 312 [i.7].

    NOTE: Cryptographic suites recommendations defined in ETSI TS 119 312 [i.7] can be superseded by national
    recommendations.
    """

    VALIDATION_DISCOURAGED_PUBLIC_KEY_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "etsi.en_319_412_2.gen-4.2.5-1.discouraged_public_key_type",
    )

    def __init__(self):
        super().__init__(
            validation=self.VALIDATION_DISCOURAGED_PUBLIC_KEY_TYPE,
            allowed_encodings=_SPKI_ALG_ID_ENCODINGS,
            pdu_class=rfc5280.AlgorithmIdentifier,
        )


_NIST_ALG_ARC = univ.ObjectIdentifier("2.16.840.1.101.3.4")
_SHA3_ARC = _NIST_ALG_ARC + (2,)

id_sha3_224 = _SHA3_ARC + (7,)
id_sha3_256 = _SHA3_ARC + (8,)
id_sha3_384 = _SHA3_ARC + (9,)
id_sha3_512 = _SHA3_ARC + (10,)


_ECDSA_WITH_SHA3_ARC = _NIST_ALG_ARC + (3,)

id_ecdsa_with_sha3_256 = _ECDSA_WITH_SHA3_ARC + (10,)
id_ecdsa_with_sha3_384 = _ECDSA_WITH_SHA3_ARC + (11,)
id_ecdsa_with_sha3_512 = _ECDSA_WITH_SHA3_ARC + (12,)


_HASH_ALG_ID_TO_HASH_LENGTH = {
    o: l / 8
    for o, l in (
        (rfc4055.id_sha224, 224),
        (rfc4055.id_sha256, 256),
        (rfc4055.id_sha384, 384),
        (rfc4055.id_sha512, 512),
        (id_sha3_224, 224),
        (id_sha3_256, 256),
        (id_sha3_384, 384),
        (id_sha3_512, 512),
    )
}


def _create_rsapss_params(hash_alg, encode_null_params):
    params = rfc4055.RSASSA_PSS_params()
    params["hashAlgorithm"]["algorithm"] = hash_alg

    if encode_null_params:
        params["hashAlgorithm"]["parameters"] = univ.Null("")

    params["maskGenAlgorithm"]["algorithm"] = rfc4055.id_mgf1

    mask_gen_alg_id = rfc5280.AlgorithmIdentifier()
    mask_gen_alg_id["algorithm"] = hash_alg

    if encode_null_params:
        mask_gen_alg_id["parameters"] = univ.Null("")

    params["maskGenAlgorithm"]["parameters"] = mask_gen_alg_id

    params["saltLength"] = _HASH_ALG_ID_TO_HASH_LENGTH[hash_alg]

    return params


_SIGNATURE_ALG_ID_ENCODINGS = (
    [
        # PKCS v1.5
        _create_alg_id_der(o, univ.Null(""))
        for o in (
            rfc4055.sha256WithRSAEncryption,
            rfc4055.sha384WithRSAEncryption,
            rfc4055.sha512WithRSAEncryption,
        )
    ]
    + [
        # RSASSA-PSS without null parameters
        _create_alg_id_der(rfc4055.id_RSASSA_PSS, _create_rsapss_params(h, False))
        for h in (
            rfc4055.id_sha224,
            rfc4055.id_sha256,
            rfc4055.id_sha384,
            rfc4055.id_sha512,
            id_sha3_224,
            id_sha3_256,
            id_sha3_384,
            id_sha3_512,
        )
    ]
    + [
        # RSASSA-PSS with null parameters
        _create_alg_id_der(rfc4055.id_RSASSA_PSS, _create_rsapss_params(h, True))
        for h in (
            rfc4055.id_sha224,
            rfc4055.id_sha256,
            rfc4055.id_sha384,
            rfc4055.id_sha512,
            id_sha3_224,
            id_sha3_256,
            id_sha3_384,
            id_sha3_512,
        )
    ]
    + [
        # ECDSA
        _create_alg_id_der(o, None)
        for o in (
            rfc5480.ecdsa_with_SHA256,
            rfc5480.ecdsa_with_SHA384,
            rfc5480.ecdsa_with_SHA512,
            id_ecdsa_with_sha3_256,
            id_ecdsa_with_sha3_384,
            id_ecdsa_with_sha3_512,
        )
    ]
    # TODO: add support for Schnorr signatures
)


class AllowedSignatureAlgorithmValidator(
    algorithm.AllowedSignatureAlgorithmEncodingValidator
):
    """
    From EN 319 412 2:

    GEN-4.2.2-1

    Signature algorithm should be selected according to ETSI TS 119 312 [i.7].

    NOTE: Cryptographic suites recommendations defined in ETSI TS 119 312 [i.7] can be superseded by national
    recommendations.
    """

    VALIDATION_DISCOURAGED_SIGNATURE_ALGORITHM = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "etsi.en_319_412_2.gen-4.2.2-1.discouraged_signature_algorithm",
    )

    def __init__(self, path: str):
        super().__init__(
            validation=self.VALIDATION_DISCOURAGED_SIGNATURE_ALGORITHM,
            allowed_encodings=_SIGNATURE_ALG_ID_ENCODINGS,
            path=path,
        )



import glob
import sys
from os import path

from pkilint import pkix
from pkilint.pkix import name, extension, crl
from tests.integration_crl import register_test

cur_dir = path.dirname(__file__)
test_dir = path.join(cur_dir, "pkix", "crl")
this_module = sys.modules[__name__]

files = glob.glob(path.join(test_dir, "*.crltest"))


for file in files:
    validator = crl.create_pkix_crl_validator_container(
        [
            pkix.create_attribute_decoder(name.ATTRIBUTE_TYPE_MAPPINGS),
            pkix.create_extension_decoder(extension.EXTENSION_MAPPINGS),
        ],
        [
            crl.create_issuer_validator_container([]),
            crl.create_validity_validator_container([]),
            crl.create_extensions_validator_container([]),
        ],
    )

    file_no_ext, _ = path.splitext(path.basename(file))

    test_name = f"test_{file_no_ext}"

    register_test(this_module, file, test_name, validator)



import ipaddress
from urllib.parse import urlparse

import publicsuffixlist
from pyasn1_alt_modules import rfc8398, rfc5280

from pkilint import validation
from pkilint.pkix import general_name


class InternalDomainNameValidator(validation.Validator):
    def __init__(
        self,
        validation_internal_domain_name_present: validation.ValidationFinding,
        *args,
        **kwargs,
    ):
        self._psl = publicsuffixlist.PublicSuffixList(accept_unknown=False)

        super().__init__(
            validations=[validation_internal_domain_name_present], **kwargs
        )

        self._validation_internal_domain_name_present = (
            validation_internal_domain_name_present
        )

    @classmethod
    def extract_domain_name(cls, node):
        return str(node.pdu)

    def validate_with_value(self, node, value):
        if self._psl.publicsuffix(value) is None:
            raise validation.ValidationFindingEncountered(
                self._validation_internal_domain_name_present,
                f'Internal domain name: "{value}"',
            )

    def validate(self, node):
        domain_name = self.extract_domain_name(node)

        return self.validate_with_value(node, domain_name)


class GeneralNameDnsNameInternalDomainNameValidator(InternalDomainNameValidator):
    def __init__(
        self,
        validation_internal_domain_name_present: validation.ValidationFinding,
        allow_onion_tld=False,
    ):
        self._allow_onion_tld = allow_onion_tld

        super().__init__(
            validation_internal_domain_name_present,
            predicate=general_name.create_generalname_type_predicate(
                general_name.GeneralNameTypeName.DNS_NAME
            ),
        )

    def validate_with_value(self, node, value):
        if len(value) == 0 and general_name.is_nameconstraints_child_node(node):
            return
        else:
            if self._allow_onion_tld and value.lower().endswith(".onion"):
                return
            else:
                return super().validate_with_value(node, value)


class UriInternalDomainNameValidator(InternalDomainNameValidator):
    def __init__(
        self,
        validation_internal_domain_name_present: validation.ValidationFinding,
        *args,
        **kwargs,
    ):
        super().__init__(validation_internal_domain_name_present, *args, **kwargs)

    def extract_domain_name(self, node):
        return urlparse(str(node.pdu)).hostname or ""


class GeneralNameUriInternalDomainNameValidator(InternalDomainNameValidator):
    def __init__(
        self, validation_internal_domain_name_present: validation.ValidationFinding
    ):
        super().__init__(
            validation_internal_domain_name_present,
            predicate=general_name.create_generalname_type_predicate(
                general_name.GeneralNameTypeName.UNIFORM_RESOURCE_IDENTIFIER
            ),
        )

    def extract_domain_name(self, node):
        if general_name.is_nameconstraints_child_node(node):
            return str(node.pdu).lstrip(".")
        else:
            return urlparse(str(node.pdu)).hostname or ""

    def validate_with_value(self, node, value):
        if len(value) == 0 and general_name.is_nameconstraints_child_node(node):
            return
        else:
            return super().validate_with_value(node, value)


class EmailAddressInternalDomainNameValidator(InternalDomainNameValidator):
    def __init__(
        self,
        validation_internal_domain_name_present: validation.ValidationFinding,
        *args,
        **kwargs,
    ):
        super().__init__(validation_internal_domain_name_present, *args, **kwargs)

    def extract_domain_name(self, node):
        parts = str(node.pdu).split("@", maxsplit=1)

        return parts[1] if len(parts) == 2 else ""


class GeneralNameRfc822NameInternalDomainNameValidator(
    EmailAddressInternalDomainNameValidator
):
    def __init__(
        self, validation_internal_domain_name_present: validation.ValidationFinding
    ):
        super().__init__(
            validation_internal_domain_name_present,
            predicate=general_name.create_generalname_type_predicate(
                general_name.GeneralNameTypeName.RFC822_NAME
            ),
        )

    def extract_domain_name(self, node):
        if general_name.is_nameconstraints_child_node(node):
            return str(node.pdu).lstrip(".")
        else:
            return super().extract_domain_name(node)

    def validate_with_value(self, node, value):
        if len(value) == 0 and general_name.is_nameconstraints_child_node(node):
            return
        else:
            super().validate_with_value(node, value)


class SmtpUtf8MailboxInternalDomainNameValidator(
    EmailAddressInternalDomainNameValidator
):
    def __init__(
        self, validation_internal_domain_name_present: validation.ValidationFinding
    ):
        super().__init__(
            validation_internal_domain_name_present, pdu_class=rfc8398.SmtpUTF8Mailbox
        )

    def extract_domain_name(self, node):
        domain_part = super().extract_domain_name(node)

        # remove ToASCII once RFC 9598 is published
        return domain_part.encode("idna").decode()


class InternalIpAddressValidator(validation.Validator):
    def __init__(
        self,
        validation_internal_ip_address_present: validation.ValidationFinding,
        *args,
        **kwargs,
    ):
        super().__init__(validations=[validation_internal_ip_address_present], **kwargs)

        self._validation_internal_ip_address_present = (
            validation_internal_ip_address_present
        )

    @classmethod
    def _extract_ip_address(cls, node):
        octets = node.pdu.asOctets()

        if len(octets) == 4:
            return ipaddress.IPv4Address(octets)
        else:
            return ipaddress.IPv6Address(octets)

    def validate(self, node):
        ip_addr = self._extract_ip_address(node)

        if not ip_addr.is_global:
            raise validation.ValidationFindingEncountered(
                self._validation_internal_ip_address_present,
                f'Internal IP address: "{ip_addr}"',
            )


class GeneralNameInternalIpAddressValidator(InternalIpAddressValidator):
    def __init__(
        self, validation_internal_ip_address_present: validation.ValidationFinding
    ):
        super().__init__(
            validation_internal_ip_address_present,
            predicate=general_name.create_generalname_type_predicate(
                general_name.GeneralNameTypeName.IP_ADDRESS
            ),
        )

    def validate(self, node):
        if general_name.is_nameconstraints_child_node(node):
            return

        super().validate(node)


def create_internal_name_validator_container(
    validation_internal_domain_name_present: validation.ValidationFinding,
    validation_internal_ip_address_present: validation.ValidationFinding,
    allow_onion_tld: bool = False,
):
    validators = [
        GeneralNameDnsNameInternalDomainNameValidator(
            validation_internal_domain_name_present, allow_onion_tld
        ),
        GeneralNameUriInternalDomainNameValidator(
            validation_internal_domain_name_present
        ),
        GeneralNameRfc822NameInternalDomainNameValidator(
            validation_internal_domain_name_present
        ),
        SmtpUtf8MailboxInternalDomainNameValidator(
            validation_internal_domain_name_present
        ),
        GeneralNameInternalIpAddressValidator(validation_internal_ip_address_present),
    ]

    return validation.ValidatorContainer(
        validators=validators, pdu_class=rfc5280.GeneralName
    )


def create_cpsuri_internal_domain_name_validator(
    validation_internal_domain_name_present: validation.ValidationFinding,
):
    return UriInternalDomainNameValidator(
        validation_internal_domain_name_present, pdu_class=rfc5280.CPSuri
    )



from pyasn1_alt_modules import rfc5280

from pkilint import validation
from pkilint.itu import bitstring
from pkilint.pkix.certificate.certificate_extension import KeyUsageBitName


# BR 7.1.2.10.7
class CaKeyUsageValidator(validation.Validator):
    VALIDATION_CA_CERT_REQUIRED_BIT_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.ca_certificate_required_ku_missing",
    )

    VALIDATION_CA_CERT_PROHIBITED_BIT_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.ca_certificate_prohibited_ku_present",
    )

    VALIDATION_CA_CERT_NO_DIG_SIG = validation.ValidationFinding(
        validation.ValidationFindingSeverity.NOTICE,
        "cabf.ca_certificate_no_digital_signature_bit",
    )

    _PROHIBITED_KUS = {str(n) for n in rfc5280.KeyUsage.namedValues} - {
        KeyUsageBitName.DIGITAL_SIGNATURE,
        KeyUsageBitName.KEY_CERT_SIGN,
        KeyUsageBitName.CRL_SIGN,
    }

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_CA_CERT_REQUIRED_BIT_MISSING,
                self.VALIDATION_CA_CERT_PROHIBITED_BIT_PRESENT,
                self.VALIDATION_CA_CERT_NO_DIG_SIG,
            ],
            pdu_class=rfc5280.KeyUsage,
        )

    def validate(self, node):
        if not bitstring.has_named_bit(node, KeyUsageBitName.KEY_CERT_SIGN):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CA_CERT_REQUIRED_BIT_MISSING, "keyCertSign not asserted"
            )
        if not bitstring.has_named_bit(node, KeyUsageBitName.CRL_SIGN):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CA_CERT_REQUIRED_BIT_MISSING, "cRLSign not asserted"
            )
        prohibited_kus_asserted = [
            k for k in self._PROHIBITED_KUS if bitstring.has_named_bit(node, k)
        ]

        if any(prohibited_kus_asserted):
            prohibited_kus_str = ", ".join(prohibited_kus_asserted)
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CA_CERT_PROHIBITED_BIT_PRESENT,
                f"Prohibited KUs present: {prohibited_kus_str}",
            )

        if not bitstring.has_named_bit(node, KeyUsageBitName.DIGITAL_SIGNATURE):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CA_CERT_NO_DIG_SIG
            )


class CaBasicConstraintsValidator(validation.Validator):
    VALIDATION_CA_BIT_NOT_SET = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "cabf.serverauth.ca_basic_constraints_ca_bit_not_set",
    )

    def __init__(self):
        super().__init__(
            validations=self.VALIDATION_CA_BIT_NOT_SET,
            pdu_class=rfc5280.BasicConstraints,
        )

    def validate(self, node):
        if not bool(node.children["cA"].pdu):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CA_BIT_NOT_SET
            )



import typing
from typing import List

from pyasn1_alt_modules import rfc5280, rfc6962

import pkilint.cabf.cabf_name
import pkilint.cabf.serverauth.serverauth_extension
import pkilint.cabf.serverauth.serverauth_name
import pkilint.cabf.serverauth.serverauth_subscriber
import pkilint.common
from pkilint import validation, cabf, etsi, document
from pkilint.cabf import cabf_key, cabf_name, cabf_extension, cabf_ca
from pkilint.cabf.serverauth import (
    serverauth_name,
    serverauth_extension,
    serverauth_constants,
    serverauth_key,
    serverauth_root,
    serverauth_ca,
    serverauth_ocsp,
    serverauth_cross_ca,
    serverauth_finding_filter,
)
from pkilint.common import alternative_name
from pkilint.pkix import name, certificate
from pkilint.pkix.certificate import certificate_validity


def _has_name_constraints(cert: certificate.RFC5280Certificate):
    return cert.get_extension_by_oid(rfc5280.id_ce_nameConstraints) is not None


def _determine_intermediate_ca_type(cert: certificate.RFC5280Certificate):
    ekus = cert.extended_key_usages

    if not ekus:
        # assume serverauth
        ekus = {rfc5280.id_kp_serverAuth}

    if rfc6962.id_kp_precertificateSigning in ekus:
        return serverauth_constants.CertificateType.PRECERT_SIGNING_CA
    elif rfc5280.id_kp_serverAuth in ekus or rfc5280.anyExtendedKeyUsage in ekus:
        if _has_name_constraints(cert):
            return serverauth_constants.CertificateType.INTERNAL_CONSTRAINED_TLS_CA
        else:
            return serverauth_constants.CertificateType.INTERNAL_UNCONSTRAINED_TLS_CA
    else:
        return serverauth_constants.CertificateType.NON_TLS_CA


def _is_ocsp_responder(cert: certificate.RFC5280Certificate):
    return rfc5280.id_kp_OCSPSigning in cert.extended_key_usages


def _is_precert(cert: certificate.RFC5280Certificate):
    return cert.get_extension_by_oid(rfc6962.id_ce_criticalPoison) is not None


def _determine_subscriber_certificate_type(cert: certificate.RFC5280Certificate):
    is_precert = _is_precert(cert)

    policy_oids = cert.policy_oids

    if serverauth_constants.ID_POLICY_EV in policy_oids:
        return (
            serverauth_constants.CertificateType.EV_PRE_CERTIFICATE
            if is_precert
            else serverauth_constants.CertificateType.EV_FINAL_CERTIFICATE
        )
    elif serverauth_constants.ID_POLICY_IV in policy_oids:
        return (
            serverauth_constants.CertificateType.IV_PRE_CERTIFICATE
            if is_precert
            else serverauth_constants.CertificateType.IV_FINAL_CERTIFICATE
        )
    elif serverauth_constants.ID_POLICY_OV in policy_oids:
        return (
            serverauth_constants.CertificateType.OV_PRE_CERTIFICATE
            if is_precert
            else serverauth_constants.CertificateType.OV_FINAL_CERTIFICATE
        )
    else:
        # "unknown" certificate types are considered to be DV Subscriber certs
        return (
            serverauth_constants.CertificateType.DV_PRE_CERTIFICATE
            if is_precert
            else serverauth_constants.CertificateType.DV_FINAL_CERTIFICATE
        )


def determine_certificate_type(
    cert: certificate.RFC5280Certificate,
) -> serverauth_constants.CertificateType:
    if cert.is_self_issued:
        return serverauth_constants.CertificateType.ROOT_CA

    if cert.is_ca:
        return _determine_intermediate_ca_type(cert)
    else:
        if _is_ocsp_responder(cert):
            return serverauth_constants.CertificateType.OCSP_RESPONDER
        else:
            return _determine_subscriber_certificate_type(cert)


def create_decoding_validators(additional_validators=None):
    if additional_validators is None:
        additional_validators = []

    additional_validators.append(
        certificate.create_qc_statements_decoder(
            pkilint.etsi.asn1.ETSI_QC_STATEMENTS_MAPPINGS
        )
    )

    return pkilint.pkix.certificate.create_decoding_validators(
        cabf.NAME_ATTRIBUTE_MAPPINGS, cabf.EXTENSION_MAPPINGS, additional_validators
    )


def create_top_level_certificate_validators(
    certificate_type: serverauth_constants.CertificateType, additional_validators=None
):
    if additional_validators is None:
        additional_validators = []

    validators = [
        serverauth_key.ServerauthAllowedSignatureAlgorithmEncodingValidator(
            path="certificate.tbsCertificate.signature"
        ),
        cabf_extension.CabfExtensionsPresenceValidator(),
    ]

    if certificate_type == serverauth_constants.CertificateType.ROOT_CA:
        validators.append(serverauth_root.RootExtensionAllowanceValidator())
    elif certificate_type in serverauth_constants.CROSS_CA_TYPES:
        validators.append(
            serverauth_cross_ca.CrossCertificateExtensionAllowanceValidator(
                certificate_type
            )
        )
    elif certificate_type in serverauth_constants.INTERMEDIATE_CERTIFICATE_TYPES:
        validators.append(
            serverauth_ca.CaCertificateExtensionAllowanceValidator(certificate_type)
        )
    elif certificate_type in serverauth_constants.SUBSCRIBER_CERTIFICATE_TYPES:
        validators.append(
            serverauth_subscriber.SubscriberExtensionAllowanceValidator(
                certificate_type
            )
        )
    elif certificate_type == serverauth_constants.CertificateType.OCSP_RESPONDER:
        validators.append(serverauth_ocsp.OcspExtensionAllowanceValidator())
    else:
        raise ValueError(f"Unsupported certificate type: {certificate_type}")

    return validators + additional_validators


def create_spki_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []

    return validation.ValidatorContainer(
        validators=[
            serverauth_key.ServerauthAllowedPublicKeyAlgorithmEncodingValidator(
                path="certificate.tbsCertificate.subjectPublicKeyInfo.algorithm"
            ),
            cabf_key.RsaKeyValidator(),
            cabf_key.EcdsaKeyValidator(),
        ]
        + additional_validators,
        path="certificate.tbsCertificate.subjectPublicKeyInfo",
    )


def create_subject_name_validators() -> List[validation.Validator]:
    return [
        serverauth_name.AttributeOrderEncodingValidator(),
        serverauth_name.AttributeValueDirectoryStringValidator(),
        serverauth_name.X520NameAttributeValueLengthValidator(),
        serverauth_name.DomainComponentAttributeValueLengthValidator(),
        serverauth_name.ValidJurisdictionCountryValidator(),
        cabf_name.ValidCountryValidator(),
        serverauth_name.ValidBusinessCategoryValidator(),
        cabf_name.CabfOrganizationIdentifierAttributeValidator(),
        serverauth_name.ServerauthRelativeDistinguishedNameContainsOneElementValidator(),
        cabf_name.SignificantAttributeValueValidator(),
        cabf_name.HTMLEntitiesValidator(),
    ]


def create_ca_name_validator_container(
    certificate_type: serverauth_constants.CertificateType, additional_validators=None
):
    if additional_validators is None:
        additional_validators = []

    validators = [
        serverauth_ca.CaRequiredSubjectAttributesValidator(certificate_type),
        serverauth_name.ServerauthDuplicateAttributeTypeValidator(certificate_type),
    ]
    validators.extend(create_subject_name_validators())

    if certificate_type == serverauth_constants.CertificateType.ROOT_CA:
        validators.append(serverauth_root.RootSubjectIssuerIdenticalEncodingValidator())

    return certificate.create_subject_validator_container(
        validators + additional_validators
    )


def create_subscriber_name_validator_container(
    certificate_type: serverauth_constants.CertificateType, additional_validators=None
):
    if additional_validators is None:
        additional_validators = []

    validators = create_subject_name_validators()
    validators.extend(
        [
            serverauth_subscriber.SubscriberCommonNameValidator(),
            serverauth_name.ServerauthDuplicateAttributeTypeValidator(certificate_type),
        ]
    )

    if certificate_type in serverauth_constants.EV_CERTIFICATE_TYPES:
        validators.extend(
            [
                serverauth_subscriber.EvSubscriberAttributeAllowanceValidator(),
                serverauth_subscriber.EvSubscriberJurisdictionPresenceValidator(),
                pkilint.cabf.serverauth.serverauth_subscriber.OrganizationIdentifierConsistentSubjectAndExtensionValidator(),
            ]
        )
    elif certificate_type in serverauth_constants.IV_CERTIFICATE_TYPES:
        validators.append(
            serverauth_subscriber.IvSubscriberAttributeAllowanceValidator()
        )
    elif certificate_type in serverauth_constants.OV_CERTIFICATE_TYPES:
        validators.append(
            serverauth_subscriber.OvSubscriberAttributeAllowanceValidator()
        )
    elif certificate_type in serverauth_constants.DV_CERTIFICATE_TYPES:
        validators.append(
            serverauth_subscriber.DvSubcriberAttributeAllowanceValidator()
        )
    else:
        raise ValueError(f"Unsupported certificate type: {certificate_type}")

    if certificate_type in serverauth_constants.IDENTITY_CERTIFICATE_TYPES:
        validators.append(
            serverauth_subscriber.IdentityCertificateStateProvinceAndLocalityPresenceValidator()
        )

    return certificate.create_subject_validator_container(
        validators + additional_validators
    )


def create_extension_validators() -> List[validation.Validator]:
    return [
        serverauth_extension.CrlDpDistributionPointCountValidator(),
        serverauth_extension.CrlDpDistributionPointNameValidator(),
        serverauth_extension.CrlDpDistributionPointValidator(),
        serverauth_extension.AuthorityInformationAccessHttpUriLocationValidator(),
        serverauth_extension.AuthorityInformationAccessUniqueLocationValidator(),
        cabf_extension.CabfAuthorityKeyIdentifierValidator(),
        alternative_name.create_internal_name_validator_container(
            cabf_name.VALIDATION_INTERNAL_DOMAIN_NAME,
            cabf_name.VALIDATION_INTERNAL_IP_ADDRESS,
            allow_onion_tld=True,
        ),
        alternative_name.create_cpsuri_internal_domain_name_validator(
            cabf_name.VALIDATION_INTERNAL_DOMAIN_NAME
        ),
        serverauth_extension.CertificatePolicyQualifierValidator(),
        cabf_extension.CpsUriHttpValidator(),
        serverauth_name.DnsNameLdhLabelSyntaxValidator(),
        serverauth_name.TorVersion3DomainNameValidator(),
    ]


def create_ca_extension_validator_container(
    certificate_type: serverauth_constants.CertificateType,
    validity_period_start_retriever: document.ValidityPeriodStartRetriever,
    additional_validators=None,
):
    if additional_validators is None:
        additional_validators = []

    validators = create_extension_validators()

    validators.extend(
        [
            serverauth_ca.CaCertificateExtensionCriticalityValidator(),
            serverauth_ca.NameConstraintsBaseTypeValidator(),
            serverauth_ca.CaCertificatePoliciesValidator(certificate_type),
            serverauth_ca.CaCertificateAuthorityInformationAccessAccessMethodPresenceValidator(),
            cabf_ca.CaKeyUsageValidator(),
            cabf_ca.CaBasicConstraintsValidator(),
        ]
    )

    if certificate_type == serverauth_constants.CertificateType.ROOT_CA:
        validators.extend(
            [
                serverauth_root.RootAkiSkiEqualityValidator(),
                serverauth_root.RootBasicConstraintsValidator(),
            ]
        )
    elif certificate_type in serverauth_constants.CROSS_CA_TYPES:
        validators.append(
            serverauth_cross_ca.CrossCertificateAllowedEkuValidator(certificate_type)
        )
    elif certificate_type == serverauth_constants.CertificateType.NON_TLS_CA:
        validators.append(serverauth_ca.NonTlsCaCertificateAllowedEkuValidator())
    elif certificate_type == serverauth_constants.CertificateType.PRECERT_SIGNING_CA:
        validators.append(
            serverauth_ca.PrecertSigningCaCertificateAllowedEkuValidator()
        )
    elif certificate_type in serverauth_constants.TLS_CA_TYPES:
        validators.append(serverauth_ca.TlsCaCertificateAllowedEkuValidator())

    if certificate_type in {
        serverauth_constants.CertificateType.EXTERNAL_UNCONSTRAINED_EV_TLS_CA,
        serverauth_constants.CertificateType.EXTERNAL_CONSTRAINED_EV_TLS_CA,
    }:
        validators.append(
            serverauth_extension.EvCpsUriPresenceValidator(
                validity_period_start_retriever
            )
        )

    if certificate_type in serverauth_constants.CONSTRAINED_TLS_CA_TYPES:
        validators.append(serverauth_ca.TlsCaTechnicallyConstrainedValidator())

    return certificate.create_extensions_validator_container(
        validators + additional_validators
    )


def create_subscriber_extension_validator_container(
    certificate_type: serverauth_constants.CertificateType,
    validity_period_start_retriever: document.ValidityPeriodStartRetriever,
    additional_validators=None,
):
    if additional_validators is None:
        additional_validators = []

    validators = create_extension_validators()

    validators.extend(
        [
            serverauth_subscriber.SubscriberEkuAllowanceValidator(),
            serverauth_subscriber.CABFOrganizationIdentifierExtensionValidator(),
            serverauth_subscriber.SubscriberExtensionCriticalityValidator(),
            serverauth_subscriber.SubscriberAuthorityInformationAccessAccessMethodPresenceValidator(),
            serverauth_subscriber.SubscriberKeyUsageValidator(),
            serverauth_subscriber.SubscriberBasicConstraintsValidator(),
            serverauth_subscriber.SubscriberPoliciesValidator(certificate_type),
        ]
    )

    if certificate_type in serverauth_constants.EV_CERTIFICATE_TYPES:
        validators.extend(
            [
                serverauth_subscriber.EvSanGeneralNameTypeValidator(),
                serverauth_extension.EvCpsUriPresenceValidator(
                    validity_period_start_retriever
                ),
                serverauth_subscriber.EvWildcardAllowanceValidator(),
            ]
        )
    else:
        validators.append(serverauth_subscriber.SubscriberSanGeneralNameTypeValidator())

    return certificate.create_extensions_validator_container(
        validators + additional_validators
    )


def create_ocsp_extension_validator_container(additional_validators=None):
    if additional_validators is None:
        additional_validators = []

    validators = create_extension_validators()

    validators.extend(
        [
            serverauth_ocsp.OcspEkuAllowanceValidator(),
            serverauth_ocsp.OcspAuthorityInformationAccessAccessMethodPresenceValidator(),
            serverauth_ocsp.OcspResponderKeyUsageValidator(),
            serverauth_ocsp.OcspBasicConstraintsValidator(),
        ]
    )

    return certificate.create_extensions_validator_container(
        validators + additional_validators
    )


def create_validity_validator_container(certificate_type, additional_validators=None):
    if additional_validators is None:
        additional_validators = []

    validators = []

    if certificate_type == serverauth_constants.CertificateType.ROOT_CA:
        validators.append(serverauth_root.RootValidityPeriodValidator())
    elif certificate_type in serverauth_constants.SUBSCRIBER_CERTIFICATE_TYPES:
        validators.append(serverauth_subscriber.SubscriberValidityPeriodValidator())

    return certificate.create_validity_validator_container(
        validators + additional_validators
    )


def create_root_ca_validators(
    validity_period_start_retriever: document.ValidityPeriodStartRetriever,
    additional_validity_validators=None,
    additional_spki_validators=None,
    additional_name_validators=None,
    additional_extension_validators=None,
    additional_top_level_validators=None,
):
    return [
        create_validity_validator_container(
            serverauth_constants.CertificateType.ROOT_CA, additional_validity_validators
        ),
        create_spki_validator_container(additional_spki_validators),
        create_ca_name_validator_container(
            serverauth_constants.CertificateType.ROOT_CA, additional_name_validators
        ),
        create_ca_extension_validator_container(
            serverauth_constants.CertificateType.ROOT_CA,
            validity_period_start_retriever,
            additional_extension_validators,
        ),
    ] + create_top_level_certificate_validators(
        serverauth_constants.CertificateType.ROOT_CA, additional_top_level_validators
    )


def create_intermediate_ca_validators(
    certificate_type: serverauth_constants.CertificateType,
    validity_period_start_retriever: document.ValidityPeriodStartRetriever,
    additional_validity_validators=None,
    additional_spki_validators=None,
    additional_name_validators=None,
    additional_extension_validators=None,
    additional_top_level_validators=None,
):
    return [
        create_validity_validator_container(
            certificate_type, additional_validity_validators
        ),
        create_spki_validator_container(additional_spki_validators),
        create_ca_name_validator_container(
            certificate_type, additional_name_validators
        ),
        create_ca_extension_validator_container(
            certificate_type,
            validity_period_start_retriever,
            additional_extension_validators,
        ),
    ] + create_top_level_certificate_validators(
        certificate_type, additional_top_level_validators
    )


def create_subscriber_validators(
    certificate_type: serverauth_constants.CertificateType,
    validity_period_start_retriever: document.ValidityPeriodStartRetriever,
    additional_validity_validators=None,
    additional_spki_validators=None,
    additional_name_validators=None,
    additional_extension_validators=None,
    additional_top_level_validators=None,
):
    return [
        create_validity_validator_container(
            certificate_type, additional_validity_validators
        ),
        create_spki_validator_container(additional_spki_validators),
        create_subscriber_name_validator_container(
            certificate_type, additional_name_validators
        ),
        create_subscriber_extension_validator_container(
            certificate_type,
            validity_period_start_retriever,
            additional_extension_validators,
        ),
    ] + create_top_level_certificate_validators(
        certificate_type, additional_top_level_validators
    )


def create_ocsp_responder_validators(
    additional_validity_validators=None,
    additional_spki_validators=None,
    additional_name_validators=None,
    additional_extension_validators=None,
    additional_top_level_validators=None,
):
    return [
        create_validity_validator_container(
            serverauth_constants.CertificateType.OCSP_RESPONDER,
            additional_validity_validators,
        ),
        create_spki_validator_container(additional_spki_validators),
        create_ca_name_validator_container(
            serverauth_constants.CertificateType.OCSP_RESPONDER,
            additional_name_validators,
        ),
        create_ocsp_extension_validator_container(additional_extension_validators),
    ] + create_top_level_certificate_validators(
        serverauth_constants.CertificateType.OCSP_RESPONDER,
        additional_top_level_validators,
    )


def create_validators(
    certificate_type: serverauth_constants.CertificateType,
    validity_period_start_retriever: typing.Optional[
        document.ValidityPeriodStartRetriever
    ] = None,
    additional_validity_validators=None,
    additional_spki_validators=None,
    additional_name_validators=None,
    additional_extension_validators=None,
    additional_top_level_validators=None,
):
    if validity_period_start_retriever is None:
        validity_period_start_retriever = (
            certificate_validity.CertificateValidityPeriodStartRetriever()
        )

    if certificate_type == serverauth_constants.CertificateType.ROOT_CA:
        return create_root_ca_validators(
            validity_period_start_retriever,
            additional_validity_validators,
            additional_spki_validators,
            additional_name_validators,
            additional_extension_validators,
            additional_top_level_validators,
        )
    elif certificate_type in serverauth_constants.INTERMEDIATE_CERTIFICATE_TYPES:
        return create_intermediate_ca_validators(
            certificate_type,
            validity_period_start_retriever,
            additional_validity_validators,
            additional_spki_validators,
            additional_name_validators,
            additional_extension_validators,
            additional_top_level_validators,
        )
    elif certificate_type in serverauth_constants.SUBSCRIBER_CERTIFICATE_TYPES:
        return create_subscriber_validators(
            certificate_type,
            validity_period_start_retriever,
            additional_validity_validators,
            additional_spki_validators,
            additional_name_validators,
            additional_extension_validators,
            additional_top_level_validators,
        )
    elif certificate_type == serverauth_constants.CertificateType.OCSP_RESPONDER:
        return create_ocsp_responder_validators(
            additional_validity_validators,
            additional_spki_validators,
            additional_name_validators,
            additional_extension_validators,
            additional_top_level_validators,
        )
    else:
        raise ValueError(f"Unsupported certificate type: {certificate_type}")


def create_serverauth_finding_filters(
    certificate_type: serverauth_constants.CertificateType, additional_filters=None
):
    if additional_filters is None:
        additional_filters = []

    filters = [
        serverauth_finding_filter.NameConstraintsCriticalityFilter(),
        serverauth_finding_filter.PolicyQualifierPresentFilter(),
        serverauth_finding_filter.DnsNameGeneralNamePreferredNameSyntaxFilter(),
    ]

    if certificate_type in serverauth_constants.SUBSCRIBER_CERTIFICATE_TYPES:
        filters += [
            serverauth_finding_filter.EndEntitySubjectKeyIdentifierMissingFilter(),
        ]

    return filters + additional_filters



import functools
from pathlib import Path

from pkilint import loader
from tests import integration_test_common

_FIXTURE_DIR = Path(__file__).parent.resolve()

_CRL_END_ASCII_ARMOR = "-----END X509 CRL-----"


def register_test(module, file, test_name, validator):
    if hasattr(module, test_name):
        raise ValueError(f"Duplicate test name in {module}: {test_name}")

    setattr(
        module,
        test_name,
        functools.partial(
            integration_test_common.run_test,
            _CRL_END_ASCII_ARMOR,
            loader.load_pem_crl,
            file,
            validator,
        ),
    )



#!/usr/bin/env python

import argparse
import sys

from pyasn1_alt_modules import rfc5280

from pkilint import loader, document, cli_util, validation, pkix, report
from pkilint.pkix import certificate, name, extension, algorithm
from pkilint.pkix.certificate import certificate_extension, certificate_key


def create_decoder_validation_container():
    decoders = [
        pkix.create_attribute_decoder(name.ATTRIBUTE_TYPE_MAPPINGS),
        pkix.create_extension_decoder(extension.EXTENSION_MAPPINGS),
        pkix.create_signature_algorithm_identifier_decoder(
            algorithm.SIGNATURE_ALGORITHM_IDENTIFIER_MAPPINGS,
            path="certificate.tbsCertificate.signature",
        ),
        certificate.create_spki_decoder(
            certificate_key.SUBJECT_PUBLIC_KEY_ALGORITHM_IDENTIFIER_MAPPINGS,
            certificate_key.SUBJECT_KEY_PARAMETER_ALGORITHM_IDENTIFIER_MAPPINGS,
        ),
    ]

    return validation.ValidatorContainer(validators=decoders, path="certificate")


def create_issuer_validation_container():
    validators = [
        name.IssuerSubjectNameBinaryEqualValidator(
            path="certificate.tbsCertificate.subject",
            subject_document_issuer_dn_path="subject:certificate.tbsCertificate.issuer",
        ),
        extension.IssuerSubjectKeyIdentifierBinaryEqualValidator(
            subject_auth_key_id_retriever=(
                lambda n: document.get_document_by_name(
                    n, "subject"
                ).get_extension_by_oid(rfc5280.id_ce_authorityKeyIdentifier)
            )
        ),
        certificate_extension.IssuerSubjectPolicyChainValidator(),
    ]

    return validation.ValidatorContainer(validators=validators)


def create_subject_validation_container():
    validators = [
        certificate_key.SubjectSignatureVerificationValidator(
            tbs_node_retriever=lambda n: n.navigate("^.tbsCertificate"),
            path="certificate.signature",
        )
    ]

    return validation.ValidatorContainer(validators=validators)


def main(cli_args=None) -> int:
    parser = argparse.ArgumentParser("RFC 5280 Issuer/Subject Certificate Chain Linter")

    subparsers = parser.add_subparsers(dest="command", required=True)
    subparsers.add_parser(
        "validations", help="Output the set of validations which this linter performs"
    )

    lint_parser = subparsers.add_parser(
        "lint", help="Lint the specified issuer and subject certificates"
    )
    cli_util.add_standard_args(lint_parser)

    lint_parser.add_argument(
        dest="issuer",
        type=argparse.FileType("rb"),
        help="The issuer certificate to lint",
    )
    lint_parser.add_argument(
        dest="subject",
        type=argparse.FileType("rb"),
        help="The subject certificate to lint",
    )

    args = parser.parse_args(cli_args)

    decoding_validation_container = create_decoder_validation_container()
    issuer_validation_container = create_issuer_validation_container()
    subject_validation_container = create_subject_validation_container()

    if args.command == "validations":
        print(
            report.report_included_validations(
                decoding_validation_container,
                issuer_validation_container,
                subject_validation_container,
            )
        )

        return 0
    else:
        doc_collection = {}

        loader_func = loader.RFC5280CertificateDocumentLoader().get_file_loader_func(
            args.document_format
        )

        try:
            issuer = loader_func(
                args.issuer, args.issuer.name, "issuer", doc_collection
            )
        except ValueError as e:
            print(f"Failed to load issuer certificate: {e}", file=sys.stderr)
            return 1

        doc_collection["issuer"] = issuer

        try:
            subject = loader_func(
                args.subject, args.subject.name, "subject", doc_collection
            )
        except ValueError as e:
            print(f"Failed to load subject certificate: {e}", file=sys.stderr)
            return 1

        doc_collection["subject"] = subject

        results = decoding_validation_container.validate(issuer.root)
        results += decoding_validation_container.validate(subject.root)
        results += issuer_validation_container.validate(issuer.root)
        results += subject_validation_container.validate(subject.root)

        print(args.format(results, args.severity))

        return cli_util.clamp_exit_code(
            report.get_findings_count(results, args.severity)
        )


if __name__ == "__main__":
    sys.exit(main())



from pyasn1.type.constraint import ValueRangeConstraint
from pyasn1_alt_modules import rfc5280

from pkilint import validation
from pkilint.pkix import extension


class CrlNumberPresenceValidator(extension.ExtensionPresenceValidator):
    def __init__(self):
        finding = validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR, "pkix.crl_number_missing"
        )

        super().__init__(
            extension_oid=rfc5280.id_ce_cRLNumber,
            validation=finding,
            pdu_class=rfc5280.CertificateList,
        )


class CrlNumberValueValidator(validation.ASN1ConstraintValidator):
    VALIDATION_FINDING_CRL_NUMBER_OOR = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "pkix.crl_number_out_of_range"
    )

    MAX_VALUE = (1 << 159) - 1

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.CRLNumber,
            validations=[self.VALIDATION_FINDING_CRL_NUMBER_OOR],
            constraint=ValueRangeConstraint(0, self.MAX_VALUE),
        )


class AuthorityKeyIdentifierPresenceValidator(extension.ExtensionPresenceValidator):
    def __init__(self):
        finding = validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            "pkix.authority_key_identifier_missing",
        )

        super().__init__(
            extension_oid=rfc5280.id_ce_authorityKeyIdentifier,
            validation=finding,
            pdu_class=rfc5280.CertificateList,
        )


class CrlNumberCriticalityValidator(extension.ExtensionCriticalityValidator):
    def __init__(self):
        finding = validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            "pkix.crl_number_extension_critical",
        )

        super().__init__(
            type_oid=rfc5280.id_ce_cRLNumber, is_critical=False, validation=finding
        )


class CrlReasonCodeValidator(validation.Validator):
    VALIDATION_UNSPECIFIED_REASON_CODE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "pkix.crl_unspecified_crl_entry_reason_code",
    )

    def __init__(self):
        super().__init__(
            pdu_class=rfc5280.CRLReason,
            validations=[self.VALIDATION_UNSPECIFIED_REASON_CODE],
        )

    def validate(self, node):
        unspecified = rfc5280.CRLReason.namedValues["unspecified"]

        if node.pdu == unspecified:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_UNSPECIFIED_REASON_CODE
            )


class CrlReasonCodeCriticalityValidator(extension.ExtensionCriticalityValidator):
    def __init__(self):
        finding = validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            "pkix.crl_reason_code_extension_critical",
        )

        super().__init__(
            type_oid=rfc5280.id_ce_cRLReasons, is_critical=False, validation=finding
        )


class CrlReasonCodeAllowlistValidator(validation.Validator):
    def __init__(
        self,
        allowed_reason_codes,
        prohibited_reason_code_validation: validation.ValidationFinding,
    ):
        self._allowed_reason_codes = allowed_reason_codes
        self._prohibited_reason_code_validation = prohibited_reason_code_validation

        super().__init__(
            pdu_class=rfc5280.CRLReason,
            validations=[self._prohibited_reason_code_validation],
        )

    def validate(self, node):
        if node.pdu not in self._allowed_reason_codes:
            raise validation.ValidationFindingEncountered(
                self._prohibited_reason_code_validation,
                f'Prohibited reason code "{node.pdu}"',
            )



from http import HTTPStatus
from importlib.metadata import version

import pytest
from fastapi.testclient import TestClient

from pkilint import report, pkix
from pkilint.cabf import serverauth
from pkilint.cabf.serverauth import serverauth_constants
from pkilint.cabf.smime import smime_constants
from pkilint.etsi import etsi_constants
from pkilint.pkix import certificate, ocsp, name, extension, crl
from pkilint.rest import app as web_app


@pytest.fixture()
def app():
    return web_app


@pytest.fixture()
def client(app):
    return TestClient(app)


def test_version(client):
    resp = client.get("/version")
    assert version("pkilint") == resp.json()["version"]


_SMBR_SPONSORED_STRICT_PEM = """-----BEGIN CERTIFICATE-----
MIIGrzCCBJegAwIBAgIUYsQ+Fan+RfQ1ToEaA+PeZh43OTEwDQYJKoZIhvcNAQEL
BQAwSDELMAkGA1UEBhMCVVMxHzAdBgNVBAoMFkZvbyBJbmR1c3RyaWVzIExpbWl0
ZWQxGDAWBgNVBAMMD0ludGVybWVkaWF0ZSBDQTAeFw0yMzA0MTkwMDAwMDBaFw0y
MzA3MTgyMzU5NTlaMIGpMSMwIQYDVQRhExpMRUlYRy1BRVlFMDBFS1hFU1ZaVVVF
QlA2NzEeMBwGA1UEChMVQWNtZSBJbmR1c3RyaWVzLCBMdGQuMQ8wDQYDVQQEDAZZ
YW1hZGExDzANBgNVBCoMBkhhbmFrbzEWMBQGA1UEAwwNWUFNQURBIEhhbmFrbzEo
MCYGCSqGSIb3DQEJARYZaGFuYWtvLnlhbWFkYUBleGFtcGxlLmNvbTCCASIwDQYJ
KoZIhvcNAQEBBQADggEPADCCAQoCggEBALD56BlDp66YkqreF8p8QPh0T+0vgUjm
yOqie30AFUj7UZKrKLVsUGCxGMzRMeWUh0xsqYm1bCcpbwn7k6A03zLpfG/wmYz9
jm9C3aWKzR+peYbxRPPRVNZ2UBdeaFSzqVIAO8Boh7hFWsKxn3svdlBOvJjslFVx
sHiSFQ3canTKD7zTVJfOgVNNr5QYhEsTrqMfnVprlVe732Ge/U6Ify1CuN2LyYfq
4b+Jyrhe4h41YwXfbAeog44+9BxZXczkPa/EkSPvTYq7qT05BeQCjXupFISidZbg
e0tu2ZLwd7Uk09z+fd1VSb58zo2gNc+gs/uPnkb3MrKoa0YBZcCPUxMCAwEAAaOC
Ai0wggIpMAwGA1UdEwEB/wQCMAAwDgYDVR0PAQH/BAQDAgeAMB8GA1UdIwQYMBaA
FNZEADJ8qA3/rE9rZu61rpssxThUMB0GA1UdDgQWBBSJGVleDvFp9cu9R+E0/OKY
zGkwkTAUBgNVHSAEDTALMAkGB2eBDAEFAwMwPQYDVR0fBDYwNDAyoDCgLoYsaHR0
cDovL2NybC5jYS5leGFtcGxlLmNvbS9pc3N1aW5nX2NhX2NybC5jcmwwSwYIKwYB
BQUHAQEEPzA9MDsGCCsGAQUFBzAChi9odHRwOi8vcmVwb3NpdG9yeS5jYS5leGFt
cGxlLmNvbS9pc3N1aW5nX2NhLmRlcjATBgNVHSUEDDAKBggrBgEFBQcDBDCB2AYD
VR0RBIHQMIHNgRloYW5ha28ueWFtYWRhQGV4YW1wbGUuY29toCYGCCsGAQUFBwgJ
oBoMGOWxseeUsOiKseWtkEBleGFtcGxlLmNvbaSBhzCBhDEjMCEGA1UEYRMaTEVJ
WEctQUVZRTAwRUtYRVNWWlVVRUJQNjcxJDAiBgNVBAoMG+OCouOCr+ODn+W3peal
reagquW8j+S8muekvjEPMA0GA1UEBAwG5bGx55SwMQ8wDQYDVQQqDAboirHlrZAx
FTATBgNVBAMMDOWxseeUsOiKseWtkDAjBgkrBgEEAYOYKgEEFhMUQUVZRTAwRUtY
RVNWWlVVRUJQNjcwEgYJKwYBBAGDmCoCBAUTA0NFTzANBgkqhkiG9w0BAQsFAAOC
AgEAE/8rQdESC9lQcnw5TnIj/DhzWqrE6S4I1F7LFgUNQB5GJUSUbnFdeExwfV+t
bjloht4frY7oJvvYyjT2t5/nv2Hrfpe95KmRhliEkEfs3ri5J/pMHa5ju1Kox49n
m8OjKkon9HMK6c7IJy2Ow1yrwDYDflVeMmZUvMr+EmUk6BdRtF40ljNwLw8xJZfh
xUzo1OjaTKu7gtYqzrFhEqijpVoxtWIBLgL7IAujPYONrxeffJ7DY6vWzBVG4C+7
iuqlrf6Y2f25yfEp0Hs9kBD26xEZUg43Zl7BxaBbJLesUk2FRD1B/N5DYZecTc7W
F1a1YUW5N15wskn8SZAXIz9xx8OThu9v7eP3qpUNaU+iaTqbjxTPGiSUYa3Jrm1y
Abh4XCOUfb4UJo23uHsNZyoLOX8lVOsesLOE/BGvlKHzT0x49uNKZq0O6lU9fxFt
iM4MRNqmNZTN9jZ1yu06cuI8nr8AEWt7Hp5OTldj5KXZFd945DqWyZHx01Uv/w5Z
U8/E3Jf1bDTbf5OLWqombrgLIWL+A/SrRvnqyLpyDv2PHJ0IgbsylDRalxeGHa1Q
3egwHqkYRzYOy3LYRphJITSGCnqRGshySonks4osE7KbXFwMEEmEWlF1S7S+VDkq
Eqpda1II90v7ae6kNwIPK+140WOhkKilZ526OHvetaZ9XUc=
-----END CERTIFICATE-----"""

_OV_FINAL_CLEAN_PEM = """-----BEGIN CERTIFICATE-----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-----END CERTIFICATE-----"""

_BAD_CERT_POLICIES_DER_PEM = """-----BEGIN CERTIFICATE-----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-----END CERTIFICATE-----"""

_CERT_WITH_TRAILER_B64 = """MIIGxDCCBKygAwIBAgIMS9Kjc1CWWKwCO6fGMA0GCSqGSIb3DQEBCwUAMEUxCzAJBgNVBAYTAlVT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"""

_OCSP_RESPONSE_B64 = """MIIDnwoBAKCCA5gwggOUBgkrBgEFBQcwAQEEggOFMIIDgTCBsKIWBBQK46D+ndQl
dpi163Lrygznvz318RgPMjAyNDA0MDIxMjM3NDdaMIGEMIGBMFkwDQYJYIZIAWUD
BAIBBQAEIDqZRndWgHOnB7/eUBhjReTNYTTbCF66odEEJfA7bwjqBCBHSmyjAfI9
yff3B4cE4cf1/JbnFnX27YguerZcP1hFQwIEAarwDYAAGA8yMDI0MDQwMzEyMzc0
N1qgERgPMjAyNDA0MTAxMjM3NDdaMAoGCCqGSM49BAMDA2kAMGYCMQDRmVmiIb4D
m9yEXiv2XtoeQi6ftpjLmlBqqRIi+3htfF/OyjdHnFuh38cQKYqqrWYCMQDKiPct
Vu7SQs587d2ZBEHQH20j5AFiGGsbI1b3+C9ZK6NIzgD6DnWlDwpSfilEarOgggJT
MIICTzCCAkswggGuoAMCAQICAQEwCgYIKoZIzj0EAwQwODELMAkGA1UEBhMCWFgx
FDASBgNVBAoMC0NlcnRzICdyIFVzMRMwEQYDVQQDDApJc3N1aW5nIENBMB4XDTI0
MDQwMjEyMzc0N1oXDTI1MDQwMjEyMzc0N1owPDELMAkGA1UEBhMCWFgxFDASBgNV
BAoMC0NlcnRzICdyIFVzMRcwFQYDVQQDDA5PQ1NQIFJlc3BvbmRlcjB2MBAGByqG
SM49AgEGBSuBBAAiA2IABFsJAbiFIyluuRnVD/oanLN0vE1AlYYoK/7KEbHZWtu1
RzSvVwv4K3IozyJrz0wl3bz+Oxo605Qw7/dj4daNLhUdkXILd5W1jaazRjlhOo+5
tajaSMZ0cRf5kZ6EJPN+yKOBhzCBhDAdBgNVHQ4EFgQUCuOg/p3UJXaYtety68oM
57899fEwHwYDVR0jBBgwFoAUjsIUCWB26pA46TmuG21SxBd9n74wDAYDVR0TAQH/
BAIwADAOBgNVHQ8BAf8EBAMCB4AwEwYDVR0lBAwwCgYIKwYBBQUHAwkwDwYJKwYB
BQUHMAEFBAIFADAKBggqhkjOPQQDBAOBigAwgYYCQRQqjNYKbGXHdGXfEVvB//i+
DiG02hraU9kGNKXeiQcPdZRajQsY/hdZPVyaykkAFVQGv29yWmTrEax+r4oZTtzG
AkFJCwtJpi7m00Qx9r/ugNWsnCFSiKUdxuvj7mg9lJtz0hexRJZKFODWJG5dUh//
Bc2w8vywgYYoduXu4QLcoP17CA=="""

_OCSP_RESPONSE_PEM = f"""-----BEGIN OCSP RESPONSE-----\n{_OCSP_RESPONSE_B64}\n-----END OCSP RESPONSE-----\n"""

_CRL_B64 = """MIIBzTCBtgIBATANBgkqhkiG9w0BAQsFADAiMQswCQYDVQQGEwJYWDETMBEGA1UE
CgwKQ1JMcyAnciBVcxcNMjQwMzI1MTg0NzAwWhcNMjQwNDAxMTg0NzAwWqBgMF4w
CgYDVR0UBAMCAQEwHwYDVR0jBBgwFoAU/NE0t8uklbG2WeoLBWIe6JqPtDowLwYD
VR0cAQH/BCUwI6AeoByGGmh0dHA6Ly9mb28uZXhhbXBsZS9jcmwuZGxshAH/MA0G
CSqGSIb3DQEBCwUAA4IBAQAN8oDSvWsg3JvUJ4MkXvczaFb72VH0J/VL5PV2cBSm
MfaVBKnUsNr1IcxT06KF8gNrDTpKqJ9fetO290swZfcPt9sEVUBVQUpdlQc3tya1
jYWmFkA3tkpqH5rBCQa3CBm1Cg8cbFBtwWgWr70NsVvfD6etjAEP9Ze+MSXnGV0p
w9EeOV07HnSD/PGQwqCiaSn5DdIDVoH8eFSGmgNLw+b4SwUjmz8PqsZwvHxJvleV
1D8cj7zdR4ywgRMjEfJZ8Bp+Tdu64Gv0doDS0iEJIshLHYkcW1okpq/tPm8kKAbD
reparePNQwhScVcDiSL73eEBIPokgG3QhohiucP5MeF1"""

_CRL_PEM = """-----BEGIN X509 CRL-----
MIIBzTCBtgIBATANBgkqhkiG9w0BAQsFADAiMQswCQYDVQQGEwJYWDETMBEGA1UE
CgwKQ1JMcyAnciBVcxcNMjQwMzI1MTg0NzAwWhcNMjQwNDAxMTg0NzAwWqBgMF4w
CgYDVR0UBAMCAQEwHwYDVR0jBBgwFoAU/NE0t8uklbG2WeoLBWIe6JqPtDowLwYD
VR0cAQH/BCUwI6AeoByGGmh0dHA6Ly9mb28uZXhhbXBsZS9jcmwuZGxshAH/MA0G
CSqGSIb3DQEBCwUAA4IBAQAN8oDSvWsg3JvUJ4MkXvczaFb72VH0J/VL5PV2cBSm
MfaVBKnUsNr1IcxT06KF8gNrDTpKqJ9fetO290swZfcPt9sEVUBVQUpdlQc3tya1
jYWmFkA3tkpqH5rBCQa3CBm1Cg8cbFBtwWgWr70NsVvfD6etjAEP9Ze+MSXnGV0p
w9EeOV07HnSD/PGQwqCiaSn5DdIDVoH8eFSGmgNLw+b4SwUjmz8PqsZwvHxJvleV
1D8cj7zdR4ywgRMjEfJZ8Bp+Tdu64Gv0doDS0iEJIshLHYkcW1okpq/tPm8kKAbD
reparePNQwhScVcDiSL73eEBIPokgG3QhohiucP5MeF1
-----END X509 CRL-----"""

_CRL_PEM_EXPECT_ERROR = """-----BEGIN X509 CRL-----
MIIBYDBKAgEBMA0GCSqGSIb3DQEBCwUAMBYxFDASBgNVBAoTC0NlcnRzICdyIFVz
Fw0yNDA0MTUxNDMzMDBaFw0yNDA1MTUxNDMzMDBaMAAwDQYJKoZIhvcNAQELBQAD
ggEBAGhq9yTTM2ZjzAxyNvXpVbOI4xQhC0L6pdjsZ13d3QFi41QvRFib13fHgcBm
+hWXFSmOT8qgMlIk74y01DBCmrVyn6mTznr49Vy9k6eBEs34F9EtQrJ5MlYNghX2
8UNNTMbQS/T7aYQuVWp4VRZsM2ZFRC1XxDdj85qraRhhc6fDGS3PS6m5vnRuZlVv
3wVB2N2zutQeZcxHDbAa68rSS3fK8jdKjC8uzbYhCvWYIc/ZUB0c+o9clwbZdkl4
eC6gxZ1/uD98+GilFUdX9JNVsi6Il1x9Upm+Oz6JZ43Ly2+yuQZu2rohZNxEzv/f
rzDRkyHn2a+5mqqc2J9asb6RFUs=
-----END X509 CRL-----"""


def _assert_validationerror_list_present(resp):
    j = resp.json()

    detail_list = j["detail"]
    assert len(detail_list) == 1

    detail_0 = detail_list[0]
    assert detail_0["loc"] == ["body"]
    assert detail_0["type"] == "value_error"
    assert "msg" in detail_0


def test_groups(client):
    resp = client.get("/certificate")
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    names = {lg["name"] for lg in j}
    assert names == {"cabf-serverauth", "cabf-smime", "etsi", "pkix"}


def test_group_no_exist(client):
    resp = client.get("/certificate/foo")
    assert resp.status_code == HTTPStatus.NOT_FOUND


def test_group_different_case(client):
    # look at what staring at those tables in 7.1 for days has done to me
    resp = client.get("/certificate/CaBf-SeRveRauTh")

    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    names = {l["name"] for l in j["linters"]}

    assert names == set(
        map(lambda c: c.to_option_str, serverauth_constants.CertificateType)
    )


def test_smime_no_cert(client):
    resp = client.post("/certificate/cabf-smime", json={})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_smime_bad_pem(client):
    resp = client.post("/certificate/cabf-smime", json={"pem": "foo"})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_smime_bad_base64(client):
    resp = client.post("/certificate/cabf-smime", json={"b64": "foo"})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_smime_detect(client):
    resp = client.post(
        "/certificate/cabf-smime", json={"pem": _SMBR_SPONSORED_STRICT_PEM}
    )

    j = resp.json()

    assert (
        j["linter"]["name"]
        == f"{smime_constants.ValidationLevel.SPONSORED}-{smime_constants.Generation.STRICT}"
    )


def test_smime_detect_bad_extension_der(client):
    resp = client.post(
        "/certificate/cabf-smime", json={"pem": _BAD_CERT_POLICIES_DER_PEM}
    )
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY

    _assert_validationerror_list_present(resp)


def test_smime_detect_not_smime(client):
    resp = client.post("/certificate/cabf-smime", json={"pem": _OV_FINAL_CLEAN_PEM})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY

    _assert_validationerror_list_present(resp)


def test_lint_smime_unknown_linter(client):
    resp = client.post(
        "/certificate/cabf-smime/lint/FOOMASTER-BAR",
        json={"pem": _SMBR_SPONSORED_STRICT_PEM},
    )
    assert resp.status_code == HTTPStatus.NOT_FOUND


def test_lint_smime(client):
    resp = client.post(
        "/certificate/cabf-smime/SPONSORED-STRICT",
        json={"pem": _SMBR_SPONSORED_STRICT_PEM},
    )
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert len(j["results"]) == 1

    finding_descriptions = j["results"][0]["finding_descriptions"]
    assert len(finding_descriptions) == 1
    finding_description = finding_descriptions[0]

    assert finding_description["severity"] == "INFO"
    assert (
        finding_description["code"] == "pkix.subject_key_identifier_method_1_identified"
    )


def test_detect_and_lint_smime(client):
    resp = client.post(
        "/certificate/cabf-smime", json={"pem": _SMBR_SPONSORED_STRICT_PEM}
    )
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert (
        j["linter"]["name"]
        == f"{smime_constants.ValidationLevel.SPONSORED.name}-{smime_constants.Generation.STRICT}"
    )

    assert len(j["results"]) == 1

    finding_descriptions = j["results"][0]["finding_descriptions"]
    assert len(finding_descriptions) == 1
    finding_description = finding_descriptions[0]

    assert finding_description["severity"] == "INFO"
    assert (
        finding_description["code"] == "pkix.subject_key_identifier_method_1_identified"
    )


def test_detect_and_lint_smime_with_tls(client):
    resp = client.post("/certificate/cabf-smime", json={"pem": _OV_FINAL_CLEAN_PEM})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY

    _assert_validationerror_list_present(resp)


def test_serverauth_no_cert(client):
    resp = client.post("/certificate/cabf-serverauth", json={})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_serverauth_bad_pem(client):
    resp = client.post("/certificate/cabf-serverauth", json={"pem": "foo"})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_serverauth_bad_base64(client):
    resp = client.post("/certificate/cabf-serverauth", json={"b64": "foo"})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_serverauth_with_bad_base64_trailer(client):
    resp = client.post(
        "/certificate/cabf-serverauth", json={"b64": _CERT_WITH_TRAILER_B64}
    )
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_serverauth_detect(client):
    resp = client.post(
        "/certificate/cabf-serverauth", json={"pem": _OV_FINAL_CLEAN_PEM}
    )

    j = resp.json()

    assert (
        j["linter"]["name"]
        == serverauth_constants.CertificateType.OV_FINAL_CERTIFICATE.to_option_str
    )


def test_serverauth_detect_not_serverauth(client):
    resp = client.post(
        "/certificate/cabf-serverauth", json={"pem": _SMBR_SPONSORED_STRICT_PEM}
    )
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert (
        j["linter"]["name"]
        == serverauth_constants.CertificateType.DV_FINAL_CERTIFICATE.to_option_str
    )


def test_lint_serverauth_unknown_linter(client):
    resp = client.post(
        "/certificate/cabf-serverauth/FOOMASTER-BAR", json={"pem": _OV_FINAL_CLEAN_PEM}
    )
    assert resp.status_code == HTTPStatus.NOT_FOUND


def test_lint_serverauth(client):
    resp = client.post(
        "/certificate/cabf-serverauth/OV-FINAL-CERTIFICATE",
        json={"pem": _OV_FINAL_CLEAN_PEM},
    )
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert len(j["results"]) == 0


def test_detect_and_lint_serverauth(client):
    resp = client.post(
        "/certificate/cabf-serverauth", json={"pem": _OV_FINAL_CLEAN_PEM}
    )
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert (
        j["linter"]["name"]
        == serverauth_constants.CertificateType.OV_FINAL_CERTIFICATE.to_option_str
    )

    assert len(j["results"]) == 0


def test_detect_and_lint_serverauth_with_smime(client):
    resp = client.post(
        "/certificate/cabf-serverauth/", json={"pem": _SMBR_SPONSORED_STRICT_PEM}
    )
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert (
        j["linter"]["name"]
        == serverauth_constants.CertificateType.DV_FINAL_CERTIFICATE.to_option_str
    )


def test_validations_list(client):
    resp = client.get("/certificate/cabf-serverauth/root-ca")
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    v = certificate.create_pkix_certificate_validator_container(
        serverauth.create_decoding_validators(),
        serverauth.create_validators(serverauth_constants.CertificateType.ROOT_CA),
    )

    for actual, expected in zip(j, report.get_included_validations(v)):
        assert actual["code"] == expected.code
        assert actual["severity"] == str(expected.severity)


def test_ocsp_pkix_validations_list(client):
    resp = client.get("/ocsp/pkix")
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    v = ocsp.create_pkix_ocsp_response_validator_container(
        [
            ocsp.create_response_decoder(),
            pkix.create_attribute_decoder(name.ATTRIBUTE_TYPE_MAPPINGS),
            pkix.create_extension_decoder(extension.EXTENSION_MAPPINGS),
        ],
        [],
    )

    for actual, expected in zip(j, report.get_included_validations(v)):
        assert actual["code"] == expected.code
        assert actual["severity"] == str(expected.severity)


def test_ocsp_pkix_lint(client):
    resp = client.post("/ocsp/pkix", json={"b64": _OCSP_RESPONSE_B64})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert len(j["results"]) == 0


def test_ocsp_pkix_lint_pem(client):
    resp = client.post("/ocsp/pkix", json={"pem": _OCSP_RESPONSE_PEM})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert len(j["results"]) == 0


def test_ocsp_pkix_lint_b64_in_pem_field(client):
    resp = client.post("/ocsp/pkix", json={"pem": _OCSP_RESPONSE_B64})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_detect_and_lint_etsi(client):
    resp = client.post("/certificate/etsi", json={"pem": _OV_FINAL_CLEAN_PEM})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert (
        j["linter"]["name"]
        == etsi_constants.CertificateType.OVCP_FINAL_CERTIFICATE.to_option_str
    )


def test_crl_pkix_validations_list(
    client, validity_additional_validators=None, doc_additional_validators=None
):
    if doc_additional_validators is None:
        doc_additional_validators = []
    if validity_additional_validators is None:
        validity_additional_validators = []

    resp = client.get("/crl/pkix/crl")
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    v = crl.create_pkix_crl_validator_container(
        [
            pkix.create_attribute_decoder(name.ATTRIBUTE_TYPE_MAPPINGS),
            pkix.create_extension_decoder(extension.EXTENSION_MAPPINGS),
        ],
        [
            crl.create_issuer_validator_container([]),
            crl.create_validity_validator_container(validity_additional_validators),
            crl.create_extensions_validator_container([]),
        ]
        + doc_additional_validators,
    )

    for actual, expected in zip(j, report.get_included_validations(v)):
        assert actual["code"] == expected.code
        assert actual["severity"] == str(expected.severity)


def test_crl_pkix_lint(client):
    resp = client.post("/crl/pkix/crl", json={"b64": _CRL_B64})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert len(j["results"]) == 0


def test_crl_pkix_lint_pem(client):
    resp = client.post("/crl/pkix/crl", json={"pem": _CRL_PEM})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert len(j["results"]) == 0


def test_crl_pkix_lint_b64_in_pem_field(client):
    resp = client.post("/crl/pkix/crl", json={"pem": _CRL_B64})
    assert resp.status_code == HTTPStatus.UNPROCESSABLE_ENTITY


def test_crl_pkix_lint_pem_with_error(client):
    resp = client.post("/crl/pkix/crl", json={"pem": _CRL_PEM_EXPECT_ERROR})
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert len(j["results"]) == 3


def test_pkix_certificate_group(client):
    resp = client.get("/certificate/pkix")
    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert j["linters"][0]["name"] == "certificate"


def test_pkix_certificate_lint(client):
    resp = client.post("/certificate/pkix", json={"pem": _OV_FINAL_CLEAN_PEM})

    assert resp.status_code == HTTPStatus.OK

    j = resp.json()

    assert j["linter"]["name"] == "certificate"

    r = j["results"][0]

    assert (
        r["finding_descriptions"][0]["code"]
        == "pkix.certificate_skid_end_entity_missing"
    )



from typing import List, Optional

from fastapi import HTTPException
from pydantic import BaseModel, Field, model_validator
from typing_extensions import Annotated

from pkilint import finding_filter, report, validation, loader, document


class Version(BaseModel):
    version: Annotated[
        str, Field(description="The version of pkilint that this server is using")
    ]


_SEVERITY_DESCRIPTION = f'The severity of the finding ({", ".join(map(str, validation.ValidationFindingSeverity))})'


class Validation(BaseModel):
    severity: Annotated[str, Field(description=_SEVERITY_DESCRIPTION)]
    code: Annotated[
        str, Field(description="The code that identifies the type of validation")
    ]


class FindingDescription(BaseModel):
    severity: Annotated[str, Field(description=_SEVERITY_DESCRIPTION)]
    code: Annotated[
        str, Field(description="The code that identifies the type of finding")
    ]
    message: Annotated[
        Optional[str],
        Field(
            description="An optional message that provides further context for the "
            "finding"
        ),
    ]


class Result(BaseModel):
    validator: Annotated[
        str,
        Field(description="The class name of the validator which returned this result"),
    ]
    node_path: Annotated[
        str,
        Field(
            description="The path in the document (or set of documents) that was validated"
        ),
    ]
    finding_descriptions: Annotated[
        List[FindingDescription],
        Field(description="The list of findings returned by the validator"),
    ]


class LintResultList(BaseModel):
    results: Annotated[
        List[Result], Field(description="The list of results returned by the linter")
    ]


class Linter(BaseModel):
    name: Annotated[str, Field(description="The name of the linter")]

    def __init__(self, validator, finding_filters=None, **kwargs):
        super().__init__(**kwargs)

        self._validator = validator
        self._finding_filters = finding_filters

    @property
    def validations(self) -> List[Validation]:
        return [
            Validation(severity=str(v.severity), code=v.code)
            for v in report.get_included_validations(self._validator)
        ]

    def lint(self, doc) -> LintResultList:
        results = self._validator.validate(doc.root)

        if self._finding_filters is not None:
            results, _ = finding_filter.filter_results(self._finding_filters, results)

        report_gen = report.ReportGeneratorJson(
            results, validation.ValidationFindingSeverity.INFO
        )
        json_str = report_gen.generate()

        return LintResultList.model_validate_json(json_str)


class LintResultListWithLinter(LintResultList):
    linter: Annotated[
        Linter,
        Field(
            description="The linter that was used for linting the specified document"
        ),
    ]


class LinterGroup(BaseModel):
    name: Annotated[str, Field(description="The name of the linter group")]
    linters: Annotated[
        List[Linter], Field(description="The set of linters in this group")
    ]

    def determine_linter(self, doc):
        pass

    def get_linter_by_name(self, name: str) -> Linter:
        try:
            return next(
                (l for l in self.linters if l.name.casefold() == name.casefold())
            )
        except StopIteration:
            raise HTTPException(404, "Linter with the specified name does not exist")


class DocumentModel(BaseModel):
    pem: Annotated[Optional[str], Field(description="A PEM-encoded ASN.1 document")] = (
        None
    )
    b64: Annotated[
        Optional[str],
        Field(description="A Base64-encoded DER representation of an ASN.1 document"),
    ] = None

    def _validate(self) -> "DocumentModel":
        if self.pem and self.b64:
            raise ValueError(
                'Cannot set both "pem" and "b64" fields; exactly one must be specified'
            )
        elif not self.pem and not self.b64:
            raise ValueError('Must set exactly one of "pem" or "b64" fields')
        else:
            return self

    def parse_document(self) -> document.Document:
        pass


class CertificateModel(DocumentModel):
    _parsed_document = None

    @model_validator(mode="after")
    def validate(self) -> "CertificateModel":
        super()._validate()

        if self.pem is not None:
            try:
                self._parsed_document = loader.load_pem_certificate(
                    self.pem, "request", "request"
                )
            except ValueError as e:
                raise ValueError("Invalid PEM text specified") from e
        else:
            try:
                self._parsed_document = loader.load_b64_certificate(
                    self.b64, "request", "request"
                )
            except ValueError as e:
                raise ValueError("Invalid Base-64 encoding specified") from e

        return self

    @property
    def parsed_document(self):
        return self._parsed_document


class OcspResponseModel(DocumentModel):
    _parsed_document = None

    @model_validator(mode="after")
    def validate(self) -> "OcspResponseModel":
        super()._validate()

        if self.pem is not None:
            try:
                self._parsed_document = loader.load_pem_ocsp_response(
                    self.pem, "request", "request"
                )
            except ValueError as e:
                raise ValueError("Invalid PEM text specified") from e
        else:
            try:
                self._parsed_document = loader.load_b64_ocsp_response(
                    self.b64, "request", "request"
                )
            except ValueError as e:
                raise ValueError("Invalid Base-64 encoding specified") from e
        return self

    @property
    def parsed_document(self):
        return self._parsed_document


class CrlModel(DocumentModel):
    _parsed_document = None

    @model_validator(mode="after")
    def validate(self) -> "CrlModel":
        super()._validate()

        if self.pem is not None:
            try:
                self._parsed_document = loader.load_pem_crl(
                    self.pem, "request", "request"
                )
            except ValueError as e:
                raise ValueError("Invalid PEM text specified") from e
        else:
            try:
                self._parsed_document = loader.load_b64_crl(
                    self.b64, "request", "request"
                )
            except ValueError as e:
                raise ValueError("Invalid Base-64 encoding specified") from e
        return self

    @property
    def parsed_document(self):
        return self._parsed_document


def create_unprocessable_entity_error_detail(
    message: str, error_type: str = "value_error"
):
    return [
        {
            "loc": ["body"],
            "type": error_type,
            "msg": message,
        }
    ]



from pyasn1.type import namedtype, char, constraint
from pyasn1.type.univ import Choice, ObjectIdentifier

id_at_businessCategory = ObjectIdentifier("2.5.4.15")

ub_business_category = 128


class X520BusinessCategory(Choice):
    pass


X520BusinessCategory.componentType = namedtype.NamedTypes(
    namedtype.NamedType(
        "teletexString",
        char.TeletexString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_business_category)
        ),
    ),
    namedtype.NamedType(
        "printableString",
        char.PrintableString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_business_category)
        ),
    ),
    namedtype.NamedType(
        "universalString",
        char.UniversalString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_business_category)
        ),
    ),
    namedtype.NamedType(
        "utf8String",
        char.UTF8String().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_business_category)
        ),
    ),
    namedtype.NamedType(
        "bmpString",
        char.BMPString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_business_category)
        ),
    ),
)

id_at_postalCode = ObjectIdentifier("2.5.4.17")

ub_postal_code = 40


class X520PostalCode(Choice):
    pass


X520PostalCode.componentType = namedtype.NamedTypes(
    namedtype.NamedType(
        "teletexString",
        char.TeletexString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_postal_code)
        ),
    ),
    namedtype.NamedType(
        "printableString",
        char.PrintableString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_postal_code)
        ),
    ),
    namedtype.NamedType(
        "universalString",
        char.UniversalString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_postal_code)
        ),
    ),
    namedtype.NamedType(
        "utf8String",
        char.UTF8String().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_postal_code)
        ),
    ),
    namedtype.NamedType(
        "bmpString",
        char.BMPString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_postal_code)
        ),
    ),
)

id_at_streetAddress = ObjectIdentifier("2.5.4.9")

ub_street_address = 128


class X520StreetAddress(Choice):
    pass


X520StreetAddress.componentType = namedtype.NamedTypes(
    namedtype.NamedType(
        "teletexString",
        char.TeletexString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_street_address)
        ),
    ),
    namedtype.NamedType(
        "printableString",
        char.PrintableString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_street_address)
        ),
    ),
    namedtype.NamedType(
        "universalString",
        char.UniversalString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_street_address)
        ),
    ),
    namedtype.NamedType(
        "utf8String",
        char.UTF8String().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_street_address)
        ),
    ),
    namedtype.NamedType(
        "bmpString",
        char.BMPString().subtype(
            subtypeSpec=constraint.ValueSizeConstraint(1, ub_street_address)
        ),
    ),
)

id_at_organizationIdentifier = ObjectIdentifier("2.5.4.97")


class X520OrganizationIdentifier(Choice):
    pass


X520OrganizationIdentifier.componentType = namedtype.NamedTypes(
    namedtype.NamedType("teletexString", char.TeletexString()),
    namedtype.NamedType("printableString", char.PrintableString()),
    namedtype.NamedType("universalString", char.UniversalString()),
    namedtype.NamedType("utf8String", char.UTF8String()),
    namedtype.NamedType("bmpString", char.BMPString()),
)

ATTRIBUTE_TYPE_MAPPINGS = {
    id_at_businessCategory: X520BusinessCategory(),
    id_at_postalCode: X520PostalCode(),
    id_at_streetAddress: X520StreetAddress(),
    id_at_organizationIdentifier: X520OrganizationIdentifier(),
}



from typing import List, Sequence, Tuple, NamedTuple

from pkilint.validation import (
    ValidationResult,
    ValidationFinding,
    ValidationFindingDescription,
)


class FindingDescriptionFilter:
    def __init__(self):
        pass

    def filter(
        self,
        result: ValidationResult,
        finding_description: ValidationFindingDescription,
    ) -> bool:
        pass


class ResultWithExcludedFindingDescriptions(NamedTuple):
    result: ValidationResult
    filter_and_finding_descriptions: List[
        Tuple[FindingDescriptionFilter, ValidationFindingDescription]
    ]


class ValidationFindingFilter(FindingDescriptionFilter):
    def __init__(self, validation_finding: ValidationFinding):
        self._validation_finding = validation_finding

        super().__init__()

    def filter(
        self,
        result: ValidationResult,
        finding_description: ValidationFindingDescription,
    ) -> bool:
        return self._validation_finding != finding_description.finding


def filter_results(
    filters: List[FindingDescriptionFilter], results: Sequence[ValidationResult]
) -> Tuple[List[ValidationResult], List[ResultWithExcludedFindingDescriptions]]:
    filtered_results = []
    results_with_exclusions = []

    for result in results:
        included_finding_descriptions = []
        filtered_finding_descriptions = []

        for finding_description in result.finding_descriptions:
            excluding_filter = next(
                (f for f in filters if not f.filter(result, finding_description)), None
            )

            if excluding_filter is None:
                included_finding_descriptions.append(finding_description)
            else:
                filtered_finding_descriptions.append(
                    (
                        excluding_filter,
                        finding_description,
                    )
                )

        if any(filtered_finding_descriptions):
            results_with_exclusions.append(
                ResultWithExcludedFindingDescriptions(
                    ValidationResult(
                        result.validator,
                        result.node,
                        [fafd[1] for fafd in filtered_finding_descriptions],
                    ),
                    filtered_finding_descriptions,
                )
            )

            filtered_results.append(
                ValidationResult(
                    result.validator, result.node, included_finding_descriptions
                )
            )
        else:
            filtered_results.append(result)

    return filtered_results, results_with_exclusions



from pyasn1.type import univ


def _OID(*components):
    output = []
    for x in tuple(components):
        if isinstance(x, univ.ObjectIdentifier):
            output.extend(list(x))
        else:
            output.append(int(x))

    return univ.ObjectIdentifier(output)


_ID_ETSI_ARC = _OID(0, 4, 0, 194121)


id_etsi_qcs_semantics_identifiers = _OID(_ID_ETSI_ARC, 1)


id_etsi_qcs_semanticsId_Natural = _OID(id_etsi_qcs_semantics_identifiers, 1)


id_etsi_qcs_SemanticsId_Legal = _OID(id_etsi_qcs_semantics_identifiers, 2)


id_etsi_qcs_semanticsId_eIDASNatural = _OID(id_etsi_qcs_semantics_identifiers, 3)


id_etsi_qcs_SemanticsId_eIDASLegal = _OID(id_etsi_qcs_semantics_identifiers, 4)


id_etsi_ext = _OID(_ID_ETSI_ARC, 2)


id_etsi_ext_valassured_ST_certs = _OID(id_etsi_ext, 1)


EXTENSION_MAPPINGS = {
    id_etsi_ext_valassured_ST_certs: univ.Null(),
}



from pyasn1.type import namedtype, constraint, char
from pyasn1_alt_modules import rfc5280

MAX = float("inf")


def _create_unbounded_directory_string_namedtypes():
    return namedtype.NamedTypes(
        namedtype.NamedType(
            "teletexString",
            char.TeletexString().subtype(
                subtypeSpec=constraint.ValueSizeConstraint(1, MAX)
            ),
        ),
        namedtype.NamedType(
            "printableString",
            char.PrintableString().subtype(
                subtypeSpec=constraint.ValueSizeConstraint(1, MAX)
            ),
        ),
        namedtype.NamedType(
            "universalString",
            char.UniversalString().subtype(
                subtypeSpec=constraint.ValueSizeConstraint(1, MAX)
            ),
        ),
        namedtype.NamedType(
            "utf8String",
            char.UTF8String().subtype(
                subtypeSpec=constraint.ValueSizeConstraint(1, MAX)
            ),
        ),
        namedtype.NamedType(
            "bmpString",
            char.BMPString().subtype(
                subtypeSpec=constraint.ValueSizeConstraint(1, MAX)
            ),
        ),
    )


class X520OrganizationNameUnbounded(rfc5280.X520OrganizationName):
    componentType = _create_unbounded_directory_string_namedtypes()


class X520OrganizationalUnitNameUnbounded(rfc5280.X520OrganizationalUnitName):
    componentType = _create_unbounded_directory_string_namedtypes()


class X520CommonNameUnbounded(rfc5280.X520CommonName):
    componentType = _create_unbounded_directory_string_namedtypes()


class X520PseudonymUnbounded(rfc5280.X520Pseudonym):
    componentType = _create_unbounded_directory_string_namedtypes()


UNBOUNDED_ATTRIBUTE_TYPE_MAPPINGS = {
    rfc5280.id_at_organizationName: X520OrganizationNameUnbounded(),
    rfc5280.id_at_organizationalUnitName: X520OrganizationalUnitNameUnbounded(),
    rfc5280.id_at_commonName: X520CommonNameUnbounded(),
    rfc5280.id_at_pseudonym: X520PseudonymUnbounded(),
}



import functools
import glob
import sys
from os import path

import tests.integration_test_common
from pkilint.pkix import certificate, name, extension
from tests.integration_certificate import register_test

cur_dir = path.dirname(__file__)
test_dir = path.join(cur_dir, "pkix")
this_module = sys.modules[__name__]

files = glob.glob(path.join(test_dir, "*.crttest"))

for file in files:
    validator = certificate.create_pkix_certificate_validator_container(
        certificate.create_decoding_validators(
            name.ATTRIBUTE_TYPE_MAPPINGS, extension.EXTENSION_MAPPINGS
        ),
        [
            certificate.create_issuer_validator_container([]),
            certificate.create_validity_validator_container(),
            certificate.create_subject_validator_container([]),
            certificate.create_extensions_validator_container([]),
        ],
    )

    file_no_ext, _ = path.splitext(path.basename(file))

    test_name = f"test_{file_no_ext}"

    register_test(this_module, file, test_name, validator)



from pkilint.pkix import (
    ocsp,
    create_attribute_decoder,
    create_extension_decoder,
    extension,
    name,
)
from pkilint.rest import model


def create_ocsp_response_linter():
    return model.Linter(
        validator=ocsp.create_pkix_ocsp_response_validator_container(
            [
                ocsp.create_response_decoder(),
                create_attribute_decoder(name.ATTRIBUTE_TYPE_MAPPINGS),
                create_extension_decoder(extension.EXTENSION_MAPPINGS),
            ],
            [],
        ),
        name="ocsp_linter",
    )



from pyasn1.type.univ import ObjectIdentifier


_ARC = ObjectIdentifier("0.4.0.194112.1")


id_qcp_natural = _ARC + (0,)

id_qcp_legal = _ARC + (1,)

id_qcp_natural_qscd = _ARC + (2,)

id_qcp_legal_qscd = _ARC + (3,)

id_qcp_web = _ARC + (4,)

id_qncp_web = _ARC + (5,)

id_qncp_web_gen = _ARC + (6,)

QUALIFIED_POLICY_OIDS = {
    id_qcp_natural,
    id_qcp_legal,
    id_qcp_natural_qscd,
    id_qcp_legal_qscd,
    id_qcp_web,
    id_qncp_web,
    id_qncp_web_gen,
}



#!/usr/bin/env python

import argparse
import sys

from pkilint import loader, report, cli_util, finding_filter
from pkilint.cabf import serverauth
from pkilint.cabf.serverauth import serverauth_constants
from pkilint.pkix import certificate

_CERTIFICATE_TYPE_OPTIONS = [
    str(t).replace("_", "-") for t in serverauth_constants.CertificateType
]


class ServerauthCertificateTypeAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        cert_type = serverauth_constants.CertificateType.from_option_str(values)

        setattr(namespace, self.dest, cert_type)


def main(cli_args=None) -> int:
    parser = argparse.ArgumentParser(
        description=f"CA/Browser Forum TLS Baseline Requirements v{serverauth_constants.BR_VERSION} Certificate Linter"
    )

    subparsers = parser.add_subparsers(dest="command", required=True)

    validations_parser = subparsers.add_parser(
        "validations", help="Output the set of validations which this linter performs"
    )
    validations_parser.add_argument(
        "-t",
        "--type",
        required=True,
        type=str.upper,
        action=ServerauthCertificateTypeAction,
        help="The type of certificate",
        choices=_CERTIFICATE_TYPE_OPTIONS,
    )

    lint_parser = subparsers.add_parser("lint", help="Lint the specified certificate")

    detect_options_group = lint_parser.add_mutually_exclusive_group(required=True)
    detect_options_group.add_argument(
        "-d",
        "--detect",
        action="store_true",
        help="Detect the type of certificate from reserved CA/B Forum policy "
        "OID, EKU(s), name constraints, and basic constraints.",
    )
    detect_options_group.add_argument(
        "-t",
        "--type",
        type=str.upper,
        action=ServerauthCertificateTypeAction,
        help="The type of certificate",
        choices=_CERTIFICATE_TYPE_OPTIONS,
    )
    lint_parser.add_argument(
        "-o",
        "--output",
        action="store_true",
        help="Output the type of certificate to standard error. This option may be "
        "useful when using the --detect option.",
    )
    lint_parser.add_argument(
        "-r",
        "--report-all",
        action="store_true",
        help="Report all findings without filtering "
        "any PKIX findings that are superseded by CA/Browser Forum requirements",
    )

    cli_util.add_certificate_validity_period_start_arg(lint_parser)

    cli_util.add_standard_args(lint_parser)
    lint_parser.add_argument(
        "file", type=argparse.FileType("rb"), help="The certificate to lint"
    )

    args = parser.parse_args(cli_args)

    if args.command == "validations":
        doc_validator = certificate.create_pkix_certificate_validator_container(
            serverauth.create_decoding_validators(),
            serverauth.create_validators(args.type),
        )

        print(report.report_included_validations(doc_validator))

        return 0
    else:
        try:
            cert = loader.RFC5280CertificateDocumentLoader().get_file_loader_func(
                args.document_format
            )(args.file, args.file.name)
        except ValueError as e:
            print(f"Failed to load certificate: {e}", file=sys.stderr)
            return 1

        if args.type:
            certificate_type = args.type
        else:
            certificate_type = serverauth.determine_certificate_type(cert)

        if args.output:
            print(certificate_type.to_option_str, file=sys.stderr)

        doc_validator = certificate.create_pkix_certificate_validator_container(
            serverauth.create_decoding_validators(),
            serverauth.create_validators(certificate_type, args.validity_period_start),
        )

        results = doc_validator.validate(cert.root)

        if not args.report_all:
            results, _ = finding_filter.filter_results(
                serverauth.create_serverauth_finding_filters(certificate_type), results
            )

        print(args.format(results, args.severity))

        return cli_util.clamp_exit_code(
            report.get_findings_count(results, args.severity)
        )


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

import argparse
import re
import sys

from pyasn1.type.univ import ObjectIdentifier

from pkilint import cli_util, loader, report
from pkilint.cabf import smime
from pkilint.cabf.smime import smime_constants
from pkilint.pkix import certificate

_CERTIFICATE_TYPES = {}
for g in smime_constants.Generation:
    for v in smime_constants.ValidationLevel:
        _CERTIFICATE_TYPES[f"{v}-{g}"] = (
            v,
            g,
        )

_CONFIG_FILE_REGEX = re.compile(
    r"(?P<oid>\d+(\.\d+)*)\D.*(?P<type>(" + "|".join(_CERTIFICATE_TYPES.keys()) + "))"
)


class SmimeCertificateTypeAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        validation_level_str, generation_str = values.split("-", maxsplit=1)

        validation_level = smime_constants.ValidationLevel[validation_level_str]
        generation = smime_constants.Generation[generation_str]

        setattr(
            namespace,
            self.dest,
            (
                validation_level,
                generation,
            ),
        )


class MappingFileAction(argparse.Action):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        mappings = {}

        for line_num, line in enumerate(values.readlines()):
            line = line.strip().upper()

            m = _CONFIG_FILE_REGEX.search(line)

            if m is None:
                raise argparse.ArgumentError(
                    self, f'Syntax error on mapping file line {line_num + 1}: "{line}"'
                )

            oid = ObjectIdentifier(m.group("oid"))
            validation_level, generation = _CERTIFICATE_TYPES[m.group("type")]

            mappings[oid] = (
                validation_level,
                generation,
            )

        setattr(namespace, self.dest, mappings)


def main(cli_args=None) -> int:
    parser = argparse.ArgumentParser(
        description=f"CA/Browser Forum S/MIME Baseline Requirements v{smime_constants.BR_VERSION} Certificate Linter"
    )

    subparsers = parser.add_subparsers(dest="command", required=True)

    validations_parser = subparsers.add_parser(
        "validations", help="Output the set of validations which this linter performs"
    )
    validations_parser.add_argument(
        "-t",
        "--type",
        required=True,
        type=str.upper,
        action=SmimeCertificateTypeAction,
        help="The type (validation level and generation) of S/MIME certificate",
        choices=list(_CERTIFICATE_TYPES.keys()),
    )

    lint_parser = subparsers.add_parser("lint", help="Lint the specified certificate")

    detect_options_group = lint_parser.add_mutually_exclusive_group(required=True)

    detect_options_group.add_argument(
        "-d",
        "--detect",
        action="store_true",
        help="Detect the type of S/MIME certificate from reserved CA/B Forum policy "
        "OID. If the type cannot be detected, then refuse to lint the certificate.",
    )
    detect_options_group.add_argument(
        "-g",
        "--guess",
        action="store_true",
        help="Detect the type of S/MIME certificate from reserved CA/B Forum policy "
        "OID. If the type cannot be detected, then use heuristics to determine "
        "the type of S/MIME certificate.",
    )
    detect_options_group.add_argument(
        "-t",
        "--type",
        type=str.upper,
        action=SmimeCertificateTypeAction,
        help="The type (validation level and generation) of S/MIME certificate",
        choices=list(_CERTIFICATE_TYPES.keys()),
    )

    lint_parser.add_argument(
        "-m",
        "--mapping",
        type=argparse.FileType("r"),
        action=MappingFileAction,
        help="Mapping file which contains OID to validation level and "
        "generation mappings. Each line of the mapping file starts with a policy OID "
        "followed by a non-numeric character and the certificate type to which the OID "
        "maps (see -t/--type option for possible values)",
        default=None,
    )
    lint_parser.add_argument(
        "-o",
        "--output",
        action="store_true",
        help="Output the type of S/MIME certificate to standard error. This option may be "
        "useful when using the --detect, --guess, or --mapping options.",
    )

    cli_util.add_certificate_validity_period_start_arg(lint_parser)

    cli_util.add_standard_args(lint_parser)

    lint_parser.add_argument(
        "file", type=argparse.FileType("rb"), help="The certificate to lint"
    )

    args = parser.parse_args(cli_args)

    if args.command == "validations":
        validation_level, generation = args.type

        doc_validator = certificate.create_pkix_certificate_validator_container(
            smime.create_decoding_validators(),
            smime.create_subscriber_validators(validation_level, generation),
        )

        print(report.report_included_validations(doc_validator))

        return 0
    else:
        try:
            cert = loader.RFC5280CertificateDocumentLoader().get_file_loader_func(
                args.document_format
            )(args.file, args.file.name)
        except ValueError as e:
            print(f"Failed to load certificate: {e}", file=sys.stderr)
            return 1

        if args.detect:
            v_g = smime.determine_validation_level_and_generation(cert, args.mapping)

            if v_g is None:
                print(
                    "Could not determine validation level and generation",
                    file=sys.stderr,
                )

                return 1
            else:
                validation_level, generation = v_g
        elif args.guess:
            validation_level, generation = smime.guess_validation_level_and_generation(
                cert, args.mapping
            )
        else:
            validation_level, generation = args.type

        if args.output:
            print(f"{validation_level}-{generation}", file=sys.stderr)

        doc_validator = certificate.create_pkix_certificate_validator_container(
            smime.create_decoding_validators(),
            smime.create_subscriber_validators(
                validation_level, generation, args.validity_period_start
            ),
        )

        results = doc_validator.validate(cert.root)

        print(args.format(results, args.severity))

        return cli_util.clamp_exit_code(
            report.get_findings_count(results, args.severity)
        )


if __name__ == "__main__":
    sys.exit(main())



import typing

from pyasn1.type import univ
from pyasn1_alt_modules import rfc5280, rfc3739

import pkilint.etsi.asn1.en_319_411_2
from pkilint import validation, oid, document, common
from pkilint.etsi import asn1 as etsi_asn1, etsi_shared
from pkilint.etsi import etsi_constants
from pkilint.etsi.asn1 import en_319_411_2
from pkilint.pkix import extension, name, Rfc2119Word
from pkilint.pkix.general_name import GeneralNameTypeName


class CertificatePoliciesCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_CERTIFICATE_POLICIES_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "etsi.en_319_412_2.gen-4.3.3-1.critical_certificate_policies_extension",
    )

    def __init__(self):
        super().__init__(
            validation=self.VALIDATION_CERTIFICATE_POLICIES_CRITICAL,
            type_oid=rfc5280.id_ce_certificatePolicies,
            is_critical=False,
        )


class SubjectAlternativeNameCriticalityValidator(
    extension.ExtensionCriticalityValidator
):
    VALIDATION_SAN_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.gen-4.3.5-1.san_extension_is_critical",
    )

    def __init__(self):
        super().__init__(
            validation=self.VALIDATION_SAN_CRITICAL,
            type_oid=rfc5280.id_ce_subjectAltName,
            is_critical=False,
        )


class IssuerAlternativeNameCriticalityValidator(
    extension.ExtensionCriticalityValidator
):
    VALIDATION_ISSUER_ALTERNATIVE_NAME_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.gen-4.3.6-1.ian_extension_is_critical",
    )

    def __init__(self):
        super().__init__(
            validation=self.VALIDATION_ISSUER_ALTERNATIVE_NAME_CRITICAL,
            type_oid=rfc5280.id_ce_issuerAltName,
            is_critical=False,
        )


class ExtendedKeyUsageCriticalityValidator(extension.ExtensionCriticalityValidator):
    VALIDATION_EXTENDED_KEY_USAGE_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.gen-4.3.10-1.eku_extension_is_critical",
    )

    def __init__(self):
        super().__init__(
            validation=self.VALIDATION_EXTENDED_KEY_USAGE_CRITICAL,
            type_oid=rfc5280.id_ce_extKeyUsage,
            is_critical=False,
        )


class CRLDistributionPointsCriticalityValidator(
    extension.ExtensionCriticalityValidator
):
    CRL_DISTRIBUTION_POINTS_CRITICAL = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.gen-4.3.11-5.crl_extension_is_critical",
    )

    def __init__(self):
        super().__init__(
            validation=self.CRL_DISTRIBUTION_POINTS_CRITICAL,
            type_oid=rfc5280.id_ce_cRLDistributionPoints,
            is_critical=False,
        )


class NaturalPersonSubjectAttributeAllowanceValidator(validation.Validator):
    """
    NAT-4.2.4-1: The subject field shall include the following attributes as specified in Recommendation ITU-T X.520:
    • countryName;
    • choice of (givenName and/or surname) or pseudonym; and
    • commonName.
    """

    VALIDATION_NATURAL_PERSON_NAME_ATTRIBUTE_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.nat-4.2.4-1.required_attribute_missing",
    )

    """
    NAT-4.2.4-4 The pseudonym attribute shall not be present if the givenName
    and surname attribute are present.
    """
    VALIDATION_MIXED_PSEUDONYM_AND_NAME_ATTRIBUTES_PRESENT = (
        validation.ValidationFinding(
            validation.ValidationFindingSeverity.ERROR,
            "etsi.en_319_412_2.nat-4.2.4-4.mixed_pseudonym_and_name_attributes_present",
        )
    )

    """
    NAT 4.2.4-3 The subject field shall not contain more than one instance of commonName and countryName
    """
    VALIDATION_PROHIBITED_DUPLICATE_ATTRIBUTE_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.nat-4.2.4-3.prohibited_duplicate_attribute_present",
    )

    _REQUIRED_ATTRIBUTES = {
        rfc5280.id_at_countryName,
        rfc5280.id_at_commonName,
    }

    _PSEUDONYM_AND_NAME_ATTRIBUTES = {
        rfc5280.id_at_givenName,
        rfc5280.id_at_surname,
        rfc5280.id_at_pseudonym,
    }

    _PROHIBITED_DUPLICATE_ATTRIBUTES = {
        rfc5280.id_at_countryName,
        rfc5280.id_at_commonName,
    }

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_NATURAL_PERSON_NAME_ATTRIBUTE_MISSING,
                self.VALIDATION_MIXED_PSEUDONYM_AND_NAME_ATTRIBUTES_PRESENT,
                self.VALIDATION_PROHIBITED_DUPLICATE_ATTRIBUTE_PRESENT,
            ],
            pdu_class=rfc5280.Name,
        )

    def validate(self, node):
        attr_counts = name.get_name_attribute_counts(node)

        attrs_present = set(attr_counts.keys())

        missing_attrs = None

        if not attrs_present.issuperset(self._REQUIRED_ATTRIBUTES):
            missing_attrs = self._REQUIRED_ATTRIBUTES - attrs_present
        elif attrs_present.isdisjoint(self._PSEUDONYM_AND_NAME_ATTRIBUTES):
            missing_attrs = self._PSEUDONYM_AND_NAME_ATTRIBUTES - attrs_present

        if missing_attrs:
            oid_str = oid.format_oids(missing_attrs)

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_NATURAL_PERSON_NAME_ATTRIBUTE_MISSING,
                f"Required attributes missing: {oid_str}",
            )

        if all((a in attrs_present for a in self._PSEUDONYM_AND_NAME_ATTRIBUTES)):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_MIXED_PSEUDONYM_AND_NAME_ATTRIBUTES_PRESENT
            )

        for a in self._PROHIBITED_DUPLICATE_ATTRIBUTES:
            if attr_counts[a] > 1:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_PROHIBITED_DUPLICATE_ATTRIBUTE_PRESENT,
                    f"Prohibited duplicate attribute present: {a}",
                )


class NaturalPersonExtensionIdentifierAllowanceValidator(
    common.ExtensionIdentifierAllowanceValidator
):
    _CODE_CLASSIFIER = "etsi.en_319_412_2"

    _ALLOWANCES = {
        # GEN-4.3.1-1
        rfc5280.id_ce_authorityKeyIdentifier: Rfc2119Word.MUST,
        # NAT-4.3.2-1
        rfc5280.id_ce_keyUsage: Rfc2119Word.MUST,
        # GEN-4.3.3-2
        rfc5280.id_ce_certificatePolicies: Rfc2119Word.MUST,
        # GEN-4.3.4-1
        rfc5280.id_ce_policyMappings: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_subjectAltName: Rfc2119Word.MAY,
        rfc5280.id_ce_issuerAltName: Rfc2119Word.MAY,
        rfc5280.id_ce_subjectDirectoryAttributes: Rfc2119Word.MAY,
        # GEN-4.3.8-1
        rfc5280.id_ce_nameConstraints: Rfc2119Word.MUST_NOT,
        # GEN-4.3.9-1
        rfc5280.id_ce_policyConstraints: Rfc2119Word.MUST_NOT,
        rfc5280.id_ce_cRLDistributionPoints: Rfc2119Word.MAY,
        # GEN-4.3.12-1
        rfc5280.id_ce_inhibitAnyPolicy: Rfc2119Word.MUST_NOT,
        # GEN-4.4.1-2
        rfc5280.id_pe_authorityInfoAccess: Rfc2119Word.MUST,
    }

    def __init__(self, certificate_type: etsi_constants.CertificateType):
        allowances = self._ALLOWANCES.copy()

        if certificate_type in etsi_constants.EU_QWAC_TYPES:
            allowances[rfc3739.id_pe_qcStatements] = Rfc2119Word.MUST
        else:
            allowances[rfc3739.id_pe_qcStatements] = Rfc2119Word.MAY

        super().__init__(allowances, self._CODE_CLASSIFIER, Rfc2119Word.MAY)


def _general_name_has_uri_prefixes(uri_prefixes_lower, general_name_node):
    gn_type, gn_value = general_name_node.child

    if gn_type == GeneralNameTypeName.UNIFORM_RESOURCE_IDENTIFIER:
        uri_lower = str(gn_value.pdu).lower()

        return any(uri_lower.startswith(p) for p in uri_prefixes_lower)

    return False


_HTTP_OR_HTTPS_PREFIXES = {"http://", "https://"}
_HTTP_OR_LDAP_PREFIXES = {"http://", "ldap://"}


def _aia_extension_has_aia_ocsp_http_uri(aia_syntax_node):
    for ad in aia_syntax_node.children.values():
        if ad.children["accessMethod"].pdu == rfc5280.id_ad_ocsp:

            location_node = ad.children["accessLocation"]

            if _general_name_has_uri_prefixes(_HTTP_OR_HTTPS_PREFIXES, location_node):
                return True

    return False


class CrlDistributionPointsExtensionPresenceValidator(validation.Validator):
    """
    GEN-4.3.11-2: If the certificate does not include any access location of an OCSP responder as specified in clause
    4.4.1, then the certificate shall include a CRL distribution point extension.
    """

    VALIDATION_CRLDP_EXTENSION_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.gen-4.3.11-2.crldp_extension_missing",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_CRLDP_EXTENSION_MISSING],
            pdu_class=rfc5280.Extensions,
        )

    def validate(self, node):
        crldp_ext_and_idx = node.document.get_extension_by_oid(
            rfc5280.id_ce_cRLDistributionPoints
        )

        if crldp_ext_and_idx:
            return

        aia_ext_and_idx = node.document.get_extension_by_oid(
            rfc5280.id_pe_authorityInfoAccess
        )

        if aia_ext_and_idx is None:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CRLDP_EXTENSION_MISSING
            )

        ext, idx = aia_ext_and_idx

        try:
            _, aia_syntax_node = ext.children["extnValue"].child
        except ValueError:
            # no decoded AIA extension. Let other validator flag that finding
            return

        if not _aia_extension_has_aia_ocsp_http_uri(aia_syntax_node):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_CRLDP_EXTENSION_MISSING
            )


class CrlDistributionPointsValidator(validation.Validator):
    """
    GEN-4.3.11-4: At least one of the present references shall use either http (http://) IETF RFC 7230-7235 [3] or ldap
    (ldap://) IETF RFC 4516 [4] scheme.
    """

    VALIDATION_CRLDP_DP_NO_REQUIRED_URI = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.gen-4.3.11-4.http_or_ldap_crldp_distribution_point_uri_missing",
    )

    def __init__(self):
        super().__init__(
            validations=[self.VALIDATION_CRLDP_DP_NO_REQUIRED_URI],
            pdu_class=rfc5280.CRLDistributionPoints,
        )

    def validate(self, node):
        for distribution_point in node.children.values():
            try:
                full_name_node = distribution_point.navigate(
                    "distributionPoint.fullName"
                )
            except document.PDUNavigationFailedError:
                continue

            if any(
                _general_name_has_uri_prefixes(_HTTP_OR_LDAP_PREFIXES, gn)
                for gn in full_name_node.children.values()
            ):
                return

        raise validation.ValidationFindingEncountered(
            self.VALIDATION_CRLDP_DP_NO_REQUIRED_URI
        )


class AuthorityInformationAccessValidator(validation.Validator):
    """
    GEN-4.4.1-3: The Authority Information Access extension shall include an accessMethod OID,
    id-ad-caIssuers, with an accessLocation value specifying at least one access location of a valid CA
    certificate of the issuing CA.
    """

    VALIDATION_AIA_CA_ISSUERS_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.gen-4.4.1-3.ca_issuers_aia_access_method_absent",
    )

    """
    GEN-4.4.1-4: At least one accessLocation shall use the http (http://) IETF RFC 7230-7235 [3] scheme or https
    (https://) IETF RFC 2818 [5] scheme.
    """
    VALIDATION_AIA_CA_ISSUERS_HTTP_URI_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.gen-4.4.1-4.ca_issuers_aia_access_method_http_uri_missing",
    )

    """
    GEN-4.4.1-6: If OCSP is supported by the issuing CA, at least one access location shall specify either the http
    (http://) IETF RFC 7230-7235 [3] or https (https://) IETF RFC 2818 [5] scheme.
    """
    VALIDATION_AIA_OCSP_HTTP_URI_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.gen-4.4.1-6.ocsp_aia_access_method_http_uri_missing",
    )

    """
    GEN-4.4.1-8: If the certificate does not include any CRL distribution point extension in accordance with clause
    4.3.11, a reference to at least one OCSP responder shall be present.
    """
    VALIDATION_AIA_OCSP_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.gen-4.4.1-8.ocsp_aia_access_method_absent",
    )

    def __init__(self):
        super().__init__(
            validations=[
                self.VALIDATION_AIA_CA_ISSUERS_MISSING,
                self.VALIDATION_AIA_CA_ISSUERS_HTTP_URI_MISSING,
                self.VALIDATION_AIA_OCSP_MISSING,
                self.VALIDATION_AIA_OCSP_HTTP_URI_MISSING,
            ],
            pdu_class=rfc5280.AuthorityInfoAccessSyntax,
        )

    @classmethod
    def _get_locations_for_method(cls, method_oid, aia_syntax_node):
        return [
            ad.children["accessLocation"]
            for ad in aia_syntax_node.children.values()
            if ad.children["accessMethod"].pdu == method_oid
        ]

    def validate(self, node):
        ca_issuers_gns = self._get_locations_for_method(rfc5280.id_ad_caIssuers, node)

        if not ca_issuers_gns:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AIA_CA_ISSUERS_MISSING
            )

        if not any(
            _general_name_has_uri_prefixes(_HTTP_OR_HTTPS_PREFIXES, gn)
            for gn in ca_issuers_gns
        ):
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_AIA_CA_ISSUERS_HTTP_URI_MISSING
            )

        # only check for OCSP if no CRL DP extension is absent
        if (
            node.document.get_extension_by_oid(rfc5280.id_ce_cRLDistributionPoints)
            is None
        ):
            ocsp_gns = self._get_locations_for_method(rfc5280.id_ad_ocsp, node)

            if not ocsp_gns:
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_AIA_OCSP_MISSING
                )

            if not any(
                _general_name_has_uri_prefixes(_HTTP_OR_HTTPS_PREFIXES, gn)
                for gn in ocsp_gns
            ):
                raise validation.ValidationFindingEncountered(
                    self.VALIDATION_AIA_OCSP_HTTP_URI_MISSING
                )


class QualifiedCertificatePoliciesValidator(validation.Validator):
    """
    QCS-5.2-1: When certificates are issued as EU Qualified Certificates, they should include, in the certificate
    policies extension, one of the certificate policy identifiers defined in clause 5.3 of ETSI EN 319 411-2
    """

    VALIDATION_RECOMMENDED_POLICY_IDENTIFIER_MISSING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "etsi.en_319_412_2.qcs-5.2-1.recommended_certificate_type_policy_identifier_missing",
    )

    VALIDATION_MULTIPLE_POLICY_IDENTIFIERS_PRESENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.qcs-5.2-1.multiple_certificate_type_policy_identifiers_present",
    )

    """
    QCS-5.2-2: Policy identifiers included in the certificate policies extension of EU Qualified Certificates shall be
    consistent with the QCStatements according to clause 5.1.
    """
    VALIDATION_MISMATCHED_POLICY_IDENTIFIER_FOR_TYPE = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.qcs-5.2-2.mismatched_policy_identifier_for_certificate_type",
    )

    # TODO: add EU qualified non-website types
    _CERTIFICATE_TYPE_SET_TO_POLICY_IDENTIFIER_MAPPINGS = [
        (etsi_constants.QEVCP_W_EIDAS_CERTIFICATE_TYPES, en_319_411_2.id_qcp_web),
        (etsi_constants.QNCP_W_OV_EIDAS_CERTIFICATE_TYPES, en_319_411_2.id_qncp_web),
        (etsi_constants.QNCP_W_IV_EIDAS_CERTIFICATE_TYPES, en_319_411_2.id_qncp_web),
        (
            etsi_constants.QNCP_W_GEN_LP_EIDAS_CERTIFICATE_TYPES,
            en_319_411_2.id_qncp_web_gen,
        ),
        (
            etsi_constants.QNCP_W_GEN_NP_EIDAS_CERTIFICATE_TYPES,
            en_319_411_2.id_qncp_web_gen,
        ),
    ]

    def __init__(self, certificate_type: etsi_constants.CertificateType):
        super().__init__(
            validations=[
                self.VALIDATION_RECOMMENDED_POLICY_IDENTIFIER_MISSING,
                self.VALIDATION_MULTIPLE_POLICY_IDENTIFIERS_PRESENT,
                self.VALIDATION_MISMATCHED_POLICY_IDENTIFIER_FOR_TYPE,
            ],
            pdu_class=rfc5280.CertificatePolicies,
        )

        self._certificate_type = certificate_type

        self._recommended_policy_oid = (
            self._get_recommended_policy_oid_for_certificate_type()
        )

    def match(self, node):
        # TODO: add support for non-QWAC types
        return super().match(node) and self._recommended_policy_oid is not None

    def _get_recommended_policy_oid_for_certificate_type(
        self,
    ) -> typing.Optional[univ.ObjectIdentifier]:
        return next(
            (
                p
                for t, p in self._CERTIFICATE_TYPE_SET_TO_POLICY_IDENTIFIER_MAPPINGS
                if self._certificate_type in t
            ),
            None,
        )

    def validate(self, node):
        policy_oids = node.document.policy_oids

        certificate_type_policy_oids = (
            policy_oids & pkilint.etsi.asn1.en_319_411_2.QUALIFIED_POLICY_OIDS
        )

        if len(certificate_type_policy_oids) > 1:
            oids = oid.format_oids(certificate_type_policy_oids)

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_MULTIPLE_POLICY_IDENTIFIERS_PRESENT,
                f"Multiple certificate type policy identifiers present: {oids}",
            )
        elif not certificate_type_policy_oids:
            raise validation.ValidationFindingEncountered(
                self.VALIDATION_RECOMMENDED_POLICY_IDENTIFIER_MISSING,
                f'Missing recommended certificate type policy identifier "{self._recommended_policy_oid}"',
            )
        elif self._recommended_policy_oid not in certificate_type_policy_oids:
            # if we're here, then there's one element in the set, so it's safe to do this
            policy_oid = next(iter(certificate_type_policy_oids))

            raise validation.ValidationFindingEncountered(
                self.VALIDATION_MISMATCHED_POLICY_IDENTIFIER_FOR_TYPE,
                f"Certificate type is {self._certificate_type} ({self._recommended_policy_oid}) but certificate "
                f'contains certificate type policy identifier "{policy_oid}"',
            )


class NaturalPersonKeyUsageValidator(etsi_shared.KeyUsageValidator):
    """
    NAT-4.3.2-2: Certificates used to validate commitment to signed content (e.g. documents, agreements and/or
    transactions) shall be limited to type A, B or F.
    """

    VALIDATION_INVALID_CONTENT_COMMITMENT_SETTING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR,
        "etsi.en_319_412_2.nat-4.3.2-1.invalid_content_commitment_setting",
    )

    """
    NAT-4.3.2-3: Of these alternatives, type A should be used (see the security note 2 below).
    """
    VALIDATION_NON_PREFERRED_CONTENT_COMMITMENT_SETTING = validation.ValidationFinding(
        validation.ValidationFindingSeverity.WARNING,
        "etsi.en_319_412_2.nat-4.3.2-3.non_preferred_content_commitment_setting",
    )

    def __init__(self, is_content_commitment_type):
        super().__init__(
            is_content_commitment_type,
            self.VALIDATION_INVALID_CONTENT_COMMITMENT_SETTING,
            self.VALIDATION_NON_PREFERRED_CONTENT_COMMITMENT_SETTING,
        )


class ExtensionsPresenceValidator(common.ExtensionsPresenceValidator):
    VALIDATION_EXTENSIONS_FIELD_ABSENT = validation.ValidationFinding(
        validation.ValidationFindingSeverity.ERROR, "etsi.extensions_field_absent"
    )

    def __init__(self):
        super().__init__(self.VALIDATION_EXTENSIONS_FIELD_ABSENT)
